from plotly.offline import plot

from .paths import PLOTLY_PATH


def save_plotly_fig(fig, name):
    div = plot(fig, include_plotlyjs=False, output_type='div', show_link=False)
    with (PLOTLY_PATH / (name + '.html')).open('w') as f:
        f.write(div)



"""Package to detect spindles, ripples, slow waves.
"""
from .spindle import DetectSpindle, merge_close, transform_signal
from .ripple import DetectRipple
from .slowwave import DetectSlowWave
from .agreement import consensus, match_events



"""Package containing Qt widgets.

The program is organized in widgets.
Roughly one widget is stored in one module.

A module should start by importing:

 - packages that are in the Python library
 - additional packages (in this order: numpy, scipy, PyQt5)
 - relative imports

Leave an empty line between each group of imports.
Then make sure that the logging works:

    from logging import getLogger
    lg = getLogger(__name__)

In code add important messages with:

    lg.info('Important message')

and messages useful for debugging:

    lg.debug('Message for debugging')
"""



"""Package containing additional functions and classes, such as:
    - exceptions
    - simulate (functions to create fake data, channels for testing purposes)

"""
from .exceptions import UnrecognizedFormat, MissingDependency
from .simulate import create_data, create_channels



from numpy.testing import assert_array_equal
from pytest import raises

from wonambi import Dataset
from wonambi.ioeeg import fieldtrip
from wonambi.utils import create_data

from .paths import (fieldtrip_file,
                    hdf5_file,
                    )


def test_write_read_fieldtrip():
    data = create_data(n_trial=1, n_chan=2)

    data.export(fieldtrip_file, export_format='fieldtrip')
    d = Dataset(fieldtrip_file)
    ftdata = d.read_data()
    assert_array_equal(data.data[0], ftdata.data[0])

    assert len(d.read_markers()) == 0


def test_write_read_fieldtrip_hdf5():
    d = Dataset(hdf5_file)
    d.read_data()


def test_wrong_variable_name():
    fieldtrip.VAR = 'unknown'
    with raises(KeyError):
        Dataset(fieldtrip_file)

    with raises(KeyError):
        Dataset(hdf5_file)



from datetime import datetime
from logging import getLogger
from numpy import around, empty
from scipy.io import loadmat, savemat

from .utils import read_hdf5_chan_name
from ..utils import MissingDependency

try:
    from h5py import File
except ImportError as err:
    File = MissingDependency(err)


lg = getLogger(__name__)
VAR = 'data'


class FieldTrip:
    """Basic class to read the data.

    Parameters
    ----------
    filename : path to file
        the name of the filename or directory

    """
    def __init__(self, filename):
        self.filename = filename

    def return_hdr(self):
        """Return the header for further use.

        Returns
        -------
        subj_id : str
            subject identification code
        start_time : datetime
            start time of the dataset
        s_freq : float
            sampling frequency
        chan_name : list of str
            list of all the channels
        n_samples : int
            number of samples in the dataset
        orig : dict
            additional information taken directly from the header

        Notes
        -----
        It only reads hdf5 matlab files and the VARiable needs to be called
        'data'

        h5py is necessary for this function
        """
        # fieldtrip does not have this information
        orig = dict()
        subj_id = str()
        start_time = datetime.fromordinal(1)  # fake

        try:
            ft_data = loadmat(self.filename, struct_as_record=True,
                              squeeze_me=True)
            if VAR not in ft_data:
                raise KeyError('Save the FieldTrip variable as ''{}'''
                               ''.format(VAR))
            ft_data = ft_data[VAR]

            s_freq = ft_data['fsample'].astype('float64').item()
            n_samples = ft_data['trial'].item().shape[1]
            chan_name = list(ft_data['label'].item())

        except NotImplementedError:

            with File(self.filename) as f:

                if VAR not in f.keys():
                    raise KeyError('Save the FieldTrip variable as ''{}'''
                                   ''.format(VAR))

                s_freq = int(f[VAR]['fsample'].value.squeeze())
                chan_name = read_hdf5_chan_name(f, f[VAR]['label'])

                n_samples = int(around(f[f[VAR]['trial'][0].item()].shape[0]))

        return subj_id, start_time, s_freq, chan_name, n_samples, orig

    def return_dat(self, chan, begsam, endsam):
        """Return the data as 2D numpy.ndarray.

        Parameters
        ----------
        chan : int or list
            index (indices) of the channels to read
        begsam : int
            index of the first sample
        endsam : int
            index of the last sample

        Returns
        -------
        numpy.ndarray
            A 2d matrix, with dimension chan X samples

        """
        TRL = 0
        try:
            ft_data = loadmat(self.filename, struct_as_record=True,
                              squeeze_me=True)
            ft_data = ft_data[VAR]

            data = ft_data['trial'].item(TRL)

        except NotImplementedError:
            from h5py import File

            with File(self.filename) as f:
                data = f[f[VAR]['trial'][TRL].item()].value.T

        return data[chan, begsam:endsam]

    def return_markers(self):
        """Return all the markers (also called triggers or events).

        Returns
        -------
        list of dict
            where each dict contains 'name' as str, 'start' and 'end' as float
            in seconds from the start of the recordings, and 'chan' as list of
            str with the channels involved (if not of relevance, it's None).

        TODO
        ----
        How to read markers / events from fieldtrip file
        """
        return []


def write_fieldtrip(data, filename):
    """Export data to FieldTrip.

    Parameters
    ----------
    data : instance of ChanTime
        data with only one trial
    filename : path to file
        file to export to (include '.mat')

    Notes
    -----
    It saves mat file using Version 6 ('-v7') because it relies on scipy.io
    functions. Therefore it cannot store data larger than 2 GB.
    """
    n_trl = data.number_of('trial')
    trial = empty(n_trl, dtype='O')
    time = empty(n_trl, dtype='O')

    for trl in range(n_trl):
        trial[trl] = data.data[trl]
        time[trl] = data.axis['time'][trl]

    ft_data = {'fsample': float(data.s_freq),
               'label': data.axis['chan'][0].astype('O'),
               'trial': trial,
               'time': time,
               'cfg': 'Converted from wonambi on ' + str(datetime.now()),
               }

    savemat(filename, {VAR: ft_data})



from wonambi import Dataset
from pytest import raises
from bidso.utils import replace_underscore

from .paths import micromed_file, bids_dir

bids_dir.mkdir(exist_ok=True, parents=True)
# TODO: we should generate the name
bids_file = bids_dir / 'sub-test_task-unknown_run-1_ieeg.eeg'


# =============================================================================
# def test_ioeeg_bids_write():
# 
#     d = Dataset(micromed_file)
#     data = d.read_data()
#     markers = d.read_markers()
#     data.export(bids_file, 'bids', markers=markers)
# 
#     d_bids = Dataset(bids_file, bids=True)
#     data_bids = d_bids.read_data()
#     markers_bids = d_bids.read_markers()
# 
#     assert len(markers) == len(markers_bids)
#     assert markers[0]['start'] == markers_bids[0]['start']
# 
#     assert data.data[0][0, 0] == data_bids.data[0][0, 0]
# =============================================================================


# =============================================================================
# def test_ioeeg_bids_multiple_freq():
# 
#     # change one channel frequency
#     with replace_underscore(bids_file, 'channels.tsv').open('r+') as f:
#         f.seek(88)
#         f.write('9')
# 
#     with raises(ValueError):
#         Dataset(bids_file, bids=True)
# =============================================================================



from collections import OrderedDict
from numpy import sum, zeros
from numpy.random import seed
from numpy.testing import assert_array_equal, assert_array_almost_equal
from pytest import raises

from wonambi.utils import create_data
from wonambi.trans import montage
from wonambi.trans.montage import compute_average_regress


seed(0)
data = create_data(attr=['chan', ])


def test_montage_01():
    with raises(TypeError):
        montage(data, ref_chan='chan00')


def test_montage_02():
    reref = montage(data, ref_chan=['chan00'])
    dat1 = reref(chan='chan00')
    assert_array_equal(dat1[0], zeros(dat1[0].shape))


def test_montage_03():
    CHAN = ('chan01', 'chan02')
    reref = montage(data, ref_chan=CHAN)
    dat1 = reref(chan=CHAN)
    assert_array_almost_equal(sum(dat1[0], axis=0), zeros((dat1[0].shape[1])),
                              decimal=6)

def test_montage_04():
    reref = montage(data, ref_chan=[])
    assert_array_equal(data.data[0], reref.data[0])


def test_montage_05():
    with raises(TypeError):
        montage(data, ref_chan=['chan00'], ref_to_avg=True)


def test_montage_06():
    reref = montage(data, ref_to_avg=True)

    dat1 = reref(trial=0)
    assert_array_almost_equal(sum(dat1, axis=0), zeros((dat1.shape[1])),
                              decimal=4)


def test_montage_regression():
    reref = montage(data, ref_to_avg=True, method='regression')
    assert_array_almost_equal(reref.data[0].mean(axis=0), 0)


def test_montage_regression_transpose():
    assert_array_equal(
        compute_average_regress(data.data[0], 0),
        compute_average_regress(data.data[0].T, 1).T
        )


def test_montage_regression_error():
    with raises(ValueError):
        compute_average_regress(data.data[0][:, :, None], 0)


def test_montage_bipolar_00():
    bipol = montage(data, bipolar=100)

    assert len(bipol.chan[0]) == 28
    assert_array_equal(
        bipol(trial=0, chan='chan00-chan01'),
        data(trial=0, chan='chan00') - data(trial=0, chan='chan01'))


def test_montage_bipolar_01():
    data_nochan = create_data()

    with raises(ValueError):
        montage(data_nochan, bipolar=100)


def test_montage_bipolar_02():
    data_wrongorder = create_data(n_trial=2, attr=['chan', ])
    data_wrongorder.axis['chan'][1] = data_wrongorder.chan[0][-1::-1]

    with raises(ValueError):
        montage(data_wrongorder, bipolar=100)


def test_montage_bipolar_03():
    data_wrongorder = create_data(attr=['chan', ])

    # you should get an error if chan is not first
    data_wrongorder.axis = OrderedDict([
        ('time', data_wrongorder.axis['time']),
        ('chan', data_wrongorder.axis['chan']),
        ])

    with raises(ValueError):
        montage(data_wrongorder, bipolar=100)



"""Module to convert from electrode to sources using linear matrices

"""
from copy import deepcopy
from functools import partial
from logging import getLogger
from multiprocessing import Pool

from numpy import (arange, asarray, atleast_2d, empty, exp, isnan, NaN,
                   nansum)
from numpy.linalg import norm
try:
    from scipy.sparse import csr_matrix as sparse
    # test if csr_matrix is the fastest approach
except ImportError:
    sparse = lambda x: x

lg = getLogger(__name__)

gauss = lambda x, s: exp(-.5 * (x ** 2 / s ** 2))


class Linear:
    """
    TODO
    ----
    both hemispheres
    """
    def __init__(self, surf, chan, threshold=20, exponent=None, std=None):
        inverse = calc_xyz2surf(surf, chan.return_xyz(), threshold=threshold,
                                exponent=exponent, std=std)
        self.inv = sparse(inverse)
        self.chan = chan.return_label()

    def __call__(self, data, parameter='chan'):
        """

        TODO
        ----
        return_xyz should follow channel order
        """
        output = deepcopy(data)  # TODO: probably not the best way
        del output.axis[parameter]
        output.axis['surf'] = empty(data.number_of('trial'), dtype='O')
        output.data = empty(data.number_of('trial'), dtype='O')

        exclude_vert = ~asarray(self.inv.sum(axis=1)).flatten().astype(bool)

        for i, one_trl in enumerate(data):
            output.axis['surf'][i] = arange(self.inv.shape[0])
            output.data[i] = self.inv.dot(data.data[i])
            output.data[i][exclude_vert, ] = NaN

        return output


def calc_xyz2surf(surf, xyz, threshold=20, exponent=None, std=None):
    """Calculate transformation matrix from xyz values to vertices.

    Parameters
    ----------
    surf : instance of wonambi.attr.Surf
        the surface of only one hemisphere.
    xyz : numpy.ndarray
        nChan x 3 matrix, with the locations in x, y, z.
    std : float
        distance in mm of the Gaussian kernel
    exponent : int
        inverse law (1-> direct inverse, 2-> inverse square, 3-> inverse cube)
    threshold : float
        distance in mm for a vertex to pick up electrode activity (if distance
        is above the threshold, one electrode does not affect a vertex).

    Returns
    -------
    numpy.ndarray
        nVertices X xyz.shape[0] matrix

    Notes
    -----
    This function is a helper when plotting onto brain surface, by creating a
    transformation matrix from the values in space (f.e. at each electrode) to
    the position of the vertices (used to show the brain surface).

    There are many ways to move from values to vertices. The crucial parameter
    is the function at which activity decreases in respect to the distance. You
    can have an inverse relationship by specifying 'exponent'. If 'exponent' is
    2, then the activity will decrease as inverse square of the distance. The
    function can be a Gaussian. With std, you specify the width of the gaussian
    kernel in mm.
    For each vertex, it uses a threshold based on the distance ('threshold'
    value, in mm). Finally, it normalizes the contribution of all the channels
    to 1, so that the sum of the coefficients for each vertex is 1.

    You can also create your own matrix (and skip calc_xyz2surf altogether) and
    pass it as attribute to the main figure.
    Because it's a loop over all the vertices, this function is pretty slow,
    but if you calculate it once, you can reuse it.
    We take advantage of multiprocessing, which speeds it up considerably.
    """
    if exponent is None and std is None:
        exponent = 1

    if exponent is not None:
        lg.debug('Vertex values based on inverse-law, with exponent ' +
                 str(exponent))
        funct = partial(calc_one_vert_inverse, xyz=xyz, exponent=exponent)
    elif std is not None:
        lg.debug('Vertex values based on gaussian, with s.d. ' + str(std))
        funct = partial(calc_one_vert_gauss, xyz=xyz, std=std)

    with Pool() as p:
        xyz2surf = p.map(funct, surf.vert)

    xyz2surf = asarray(xyz2surf)

    if exponent is not None:
        threshold_value = (1 / (threshold ** exponent))
        external_threshold_value = threshold_value
    elif std is not None:
        threshold_value = gauss(threshold, std)
        external_threshold_value = gauss(std, std) # this is around 0.607
    lg.debug('Values thresholded at ' + str(threshold_value))

    xyz2surf[xyz2surf < threshold_value] = NaN

    # here we deal with vertices that are within the threshold value but far
    # from a single electrodes, so those remain empty
    sumval = nansum(xyz2surf, axis=1)
    sumval[sumval < external_threshold_value] = NaN

    # normalize by the number of electrodes
    xyz2surf /= atleast_2d(sumval).T
    xyz2surf[isnan(xyz2surf)] = 0

    return xyz2surf


def calc_one_vert_inverse(one_vert, xyz=None, exponent=None):
    """Calculate how many electrodes influence one vertex, using the inverse
    function.

    Parameters
    ----------
    one_vert : ndarray
        vector of xyz position of a vertex
    xyz : ndarray
        nChan X 3 with the position of all the channels
    exponent : int
        inverse law (1-> direct inverse, 2-> inverse square, 3-> inverse cube)

    Returns
    -------
    ndarray
        one vector with values for one vertex
    """
    trans = empty(xyz.shape[0])
    for i, one_xyz in enumerate(xyz):
        trans[i] = 1 / (norm(one_vert - one_xyz) ** exponent)
    return trans


def calc_one_vert_gauss(one_vert, xyz=None, std=None):
    """Calculate how many electrodes influence one vertex, using a Gaussian
    function.

    Parameters
    ----------
    one_vert : ndarray
        vector of xyz position of a vertex
    xyz : ndarray
        nChan X 3 with the position of all the channels
    std : float
        distance in mm of the Gaussian kernel

    Returns
    -------
    ndarray
        one vector with values for one vertex
    """
    trans = empty(xyz.shape[0])
    for i, one_xyz in enumerate(xyz):
        trans[i] = gauss(norm(one_vert - one_xyz), std)
    return trans



from wonambi import Dataset

from .paths import bci2000_file


def test_bci2000_data():
    assert bci2000_file.exists()

    d = Dataset(bci2000_file)
    assert len(d.read_markers()) == 0

    data = d.read_data()
    assert data.data[0][0, 0] == 179.702



from numpy import isnan

from wonambi import Dataset

from .paths import axon_abf_file


d = Dataset(axon_abf_file)


def test_abf_read():
    assert len(d.header['chan_name']) == 1
    assert d.header['start_time'].minute == 47

    data = d.read_data(begtime=1, endtime=2)

    assert data.data[0][0, 0] == 2.1972655922581912

    markers = d.read_markers()
    assert len(markers) == 0


def test_abf_boundary():
    data = d.read_data(begsam=-10, endsam=5)
    assert isnan(data.data[0][0, :10]).all()

    n_smp = d.header['n_samples']
    data = d.read_data(begsam=n_smp - 2, endsam=n_smp + 10)
    assert isnan(data.data[0][0, 2:]).all()



"""Module reads and writes header and data for EDF data.
"""
from logging import getLogger

from datetime import datetime, timedelta, time, date
from pathlib import Path
from re import findall, finditer
from struct import pack

from numpy import (abs,
                   asarray,
                   empty,
                   fromfile,
                   iinfo,
                   ones,
                   max,
                   NaN,
                   newaxis,
                   repeat,
                   )

from .utils import decode, _select_blocks, DEFAULT_DATETIME

lg = getLogger(__name__)

EDF_FORMAT = 'int16'  # by definition
edf_iinfo = iinfo(EDF_FORMAT)
N_BYTES = edf_iinfo.dtype.itemsize
DIGITAL_MAX = edf_iinfo.max
DIGITAL_MIN = -1 * edf_iinfo.max  # so that digital 0 = physical 0

ANNOT_NAME = 'EDF Annotations'
PATTERN = b'(?P<onset>[+\-]\d+(?:\.\d*)?)(?:\x15(?P<duration>\d+(?:\.\d*)?))?(\x14(?P<annotation>[^\x00]*))?(?:\x14\x00)'


class Edf:
    """Provide class EDF, which can be used to read the header and the data.

    Parameters
    ----------
    edffile : str
        Full path for the EDF file

    Attributes
    ----------
    hdr : dict
        header taken from EDF file
    """
    def __init__(self, edffile):
        self.filename = Path(edffile)
        self._read_hdr()

    def _read_hdr(self):
        """Read header from EDF file.

        It only reads the header for internal purposes and adds a hdr.
        """
        with self.filename.open('rb') as f:

            hdr = {}
            assert f.tell() == 0
            assert f.read(8) == b'0       '

            # recording info
            hdr['subject_id'] = decode(f.read(80)).strip()
            hdr['recording_id'] = decode(f.read(80)).strip()

            # parse timestamp
            date_str = decode(f.read(8)).strip()

            if date_str == '':
                edf_date = DEFAULT_DATETIME.date()
            else:
                (day, month, year) = [int(x) for x in findall('(\d+)', date_str)]
                # Y2K: cutoff is 1985
                if year >= 85:
                    year += 1900
                else:
                    year += 2000
                edf_date = date(year, month, day)

            time_str = decode(f.read(8)).strip()
            if time_str == '':
                edf_time = DEFAULT_DATETIME.time()
            else:
                (hour, minute, day) = [int(x) for x in findall('(\d+)', time_str)]

                edf_time = time(hour, minute, day)

            hdr['start_time'] = datetime.combine(edf_date, edf_time)

            # misc
            hdr['header_n_bytes'] = int(f.read(8))
            f.seek(44, 1)  # reserved for EDF+
            hdr['n_records'] = int(f.read(8))
            hdr['record_length'] = float(f.read(8))  # in seconds
            nchannels = hdr['n_channels'] = int(f.read(4))

            # read channel info
            channels = range(hdr['n_channels'])
            hdr['label'] = [decode(f.read(16)).strip() for n in
                            channels]
            hdr['transducer'] = [decode(f.read(80)).strip()
                                 for n in channels]
            hdr['physical_dim'] = [decode(f.read(8)).strip() for n in
                                   channels]
            hdr['physical_min'] = [float(f.read(8)) for n in channels]
            hdr['physical_max'] = [float(f.read(8)) for n in channels]
            hdr['digital_min'] = [float(f.read(8)) for n in channels]
            hdr['digital_max'] = [float(f.read(8)) for n in channels]
            hdr['prefiltering'] = [decode(f.read(80)).strip()
                                   for n in channels]
            hdr['n_samples_per_record'] = [int(f.read(8)) for n in channels]
            f.seek(32 * nchannels, 1)  # reserved

            assert f.tell() == hdr['header_n_bytes']

            self.hdr = hdr

    def return_hdr(self):
        """Return the header for further use.

        Returns
        -------
        subj_id : str
            subject identification code
        start_time : datetime
            start time of the dataset
        s_freq : float
            sampling frequency
        chan_name : list of str
            list of all the channels
        n_samples : int
            number of samples in the dataset
        orig : dict
            additional information taken directly from the header

        Notes
        -----
        EDF+ accepts multiple frequency rates for different channels. Here, we
        use only the highest sampling frequency (normally used for EEG and MEG
        signals), and we UPSAMPLE all the other channels.
        """
        try:
            self.i_annot = self.hdr['label'].index(ANNOT_NAME)
        except ValueError:
            self.i_annot = None

        self.smp_in_blk = sum(self.hdr['n_samples_per_record'])

        self.max_smp = max(self.hdr['n_samples_per_record'])
        n_blocks = self.hdr['n_records']
        self.blocks = ones(n_blocks, dtype='int') * self.max_smp

        self.dig_min = asarray(self.hdr['digital_min'])
        self.phys_min = asarray(self.hdr['physical_min'])
        phys_range = asarray(self.hdr['physical_max']) - self.phys_min
        dig_range = asarray(self.hdr['digital_max']) - self.dig_min
        assert all(phys_range > 0)
        assert all(dig_range > 0)
        self.gain = phys_range / dig_range

        subj_id = self.hdr['subject_id']
        start_time = self.hdr['start_time']
        s_freq = self.max_smp / self.hdr['record_length']
        chan_name = [label for label in self.hdr['label'] if label != ANNOT_NAME]
        n_samples = self.max_smp * self.hdr['n_records']

        return subj_id, start_time, s_freq, chan_name, n_samples, self.hdr

    def return_dat(self, chan, begsam, endsam):
        """Read data from an EDF file.

        Reads channel by channel, and adjusts the values by calibration.

        Parameters
        ----------
        chan : list of int
            index (indices) of the channels to read
        begsam : int
            index of the first sample
        endsam : int
            index of the last sample

        Returns
        -------
        numpy.ndarray
            A 2d matrix, where the first dimension is the channels and the
            second dimension are the samples.
        """
        assert begsam < endsam

        dat = empty((len(chan), endsam - begsam))
        dat.fill(NaN)

        with self.filename.open('rb') as f:

            for i_dat, blk, i_blk in _select_blocks(self.blocks, begsam, endsam):
                dat_in_rec = self._read_record(f, blk, chan)
                dat[:, i_dat[0]:i_dat[1]] = dat_in_rec[:, i_blk[0]:i_blk[1]]

        # calibration
        dat = ((dat.astype('float64') - self.dig_min[chan, newaxis]) *
               self.gain[chan, newaxis] + self.phys_min[chan, newaxis])

        return dat

    def _read_record(self, f, blk, chans):
        """Read raw data from a single EDF channel.

        Parameters
        ----------
        i_chan : int
            index of the channel to read
        begsam : int
            index of the first sample
        endsam : int
            index of the last sample

        Returns
        -------
        numpy.ndarray
            A vector with the data as written on file, in 16-bit precision
        """
        dat_in_rec = empty((len(chans), self.max_smp))

        i_ch_in_dat = 0
        for i_ch in chans:
            offset, n_smp_per_chan = self._offset(blk, i_ch)

            f.seek(offset)
            x = fromfile(f, count=n_smp_per_chan, dtype=EDF_FORMAT)

            ratio = int(self.max_smp / n_smp_per_chan)
            dat_in_rec[i_ch_in_dat, :] = repeat(x, ratio)
            i_ch_in_dat += 1

        return dat_in_rec

    def _offset(self, blk, i_ch):
        ch_in_rec = sum(self.hdr['n_samples_per_record'][:i_ch])
        n_smp_per_chan = self.hdr['n_samples_per_record'][i_ch]
        offset_in_blk = self.smp_in_blk * blk + ch_in_rec
        offset = self.hdr['header_n_bytes'] + offset_in_blk * N_BYTES

        return offset, n_smp_per_chan

    def return_markers(self):
        """"""
        if self.i_annot is None:
            return []

        annotations = []
        with self.filename.open('rb') as f:
            for blk in range(self.hdr['n_records']):
                offset, n_smp_per_chan = self._offset(blk, self.i_annot)
                f.seek(offset)
                annotations.extend(_read_tal(f.read(n_smp_per_chan)))

        markers = []
        for annot in annotations:
            for name in annot['annotation']:
                m = {'name': name,
                     'start': annot['onset'],
                     'end': annot['onset'] + annot['dur'],
                     'chan': None,
                     }
                markers.append(m)

        return markers


def write_edf(data, filename, subj_id='X X X X', physical_max=1000):
    """Export data to FieldTrip.

    Parameters
    ----------
    data : instance of ChanTime
        data with only one trial
    filename : path to file
        file to export to (include '.mat')
    subj_id : str
        subject id
    physical_max : int
        values above this parameter will be considered saturated (and also
        those that are too negative). This parameter defines the precision.

    Notes
    -----
    Data is always recorded as 2 Byte int (which is 'int16'), so precision is
    limited. You can control the precision with physical_max. To get the
    precision:

    >>> precision = physical_max / DIGITAL_MAX

    where DIGITAL_MAX is 32767.
    """
    if data.start_time is None:
        raise ValueError('Data should contain a valid start_time (as datetime)')
    start_time = data.start_time + timedelta(seconds=data.axis['time'][0][0])

    if physical_max is None:
        physical_max = max(abs(data.data[0]))

    precision = physical_max / DIGITAL_MAX
    lg.info('Data exported to EDF will have precision ' + str(precision))

    physical_min = -1 * physical_max
    dat = data.data[0] / physical_max * DIGITAL_MAX
    dat = dat.astype(EDF_FORMAT)
    dat[dat > DIGITAL_MAX] = DIGITAL_MAX
    dat[dat < DIGITAL_MIN] = DIGITAL_MIN

    with open(filename, 'wb') as f:
        f.write('{:<8}'.format(0).encode('ascii'))
        f.write('{:<80}'.format(subj_id).encode('ascii'))  # subject_id
        f.write('{:<80}'.format('Startdate X X X X').encode('ascii'))
        f.write(start_time.strftime('%d.%m.%y').encode('ascii'))
        f.write(start_time.strftime('%H.%M.%S').encode('ascii'))

        n_smp = data.data[0].shape[1]
        s_freq = int(data.s_freq)
        n_records = n_smp // s_freq  # floor
        record_length = 1
        n_channels = data.number_of('chan')[0]

        header_n_bytes = 256 + 256 * n_channels
        f.write('{:<8d}'.format(header_n_bytes).encode('ascii'))
        f.write((' ' * 44).encode('ascii'))  # reserved for EDF+

        f.write('{:<8}'.format(n_records).encode('ascii'))
        f.write('{:<8d}'.format(record_length).encode('ascii'))
        f.write('{:<4}'.format(n_channels).encode('ascii'))

        for chan in data.axis['chan'][0]:
            f.write('{:<16}'.format(chan).encode('ascii'))  # label
        for _ in range(n_channels):
            f.write(('{:<80}').format('').encode('ascii'))  # tranducer
        for _ in range(n_channels):
            f.write('{:<8}'.format('uV').encode('ascii'))  # physical_dim
        for _ in range(n_channels):
            f.write('{:<8}'.format(physical_min).encode('ascii'))
        for _ in range(n_channels):
            f.write('{:<8}'.format(physical_max).encode('ascii'))
        for _ in range(n_channels):
            f.write('{:<8}'.format(DIGITAL_MIN).encode('ascii'))
        for _ in range(n_channels):
            f.write('{:<8}'.format(DIGITAL_MAX).encode('ascii'))
        for _ in range(n_channels):
            f.write('{:<80}'.format('').encode('ascii'))  # prefiltering
        for _ in range(n_channels):
            f.write('{:<8d}'.format(s_freq).encode('ascii'))  # n_smp in record
        for _ in range(n_channels):
            f.write((' ' * 32).encode('ascii'))

        length_record = s_freq * n_channels  # length of one record
        for i in range(n_records):
            i0 = i * s_freq
            i1 = i0 + s_freq
            x = dat[:, i0:i1].flatten(order='C')  # assumes it's ChanTimeData
            f.write(pack('<' + 'h' * length_record, *x))


def _read_tal(rawbytes):
    """Read TAL (Time-stamped Annotations Lists) using regex

    Parameters
    ----------
    rawbytes : bytes
        raw information from file

    Returns
    -------
    annotation : list of dict
        where each dict contains onset, duration, and list with the annotations
    """
    annotations = []

    for m in finditer(PATTERN, rawbytes):
        d = m.groupdict()
        annot = {'onset': float(decode(d['onset'])),
                 'dur': float(decode(d['duration'])) if d['duration'] else 0.,
                 'annotation': [decode(a) for a in d['annotation'].split(b'\x14') if a],
                 }

        if annot['annotation']:
            annotations.append(annot)

    return annotations


def remove_datetime(filename):
    """Remove datetime from filename by overwriting the date / time info.

    Parameters
    ----------
    filename : Path
        path to edf file

    Notes
    -----
    It modifies the file.

    TODO
    ----
    This function might be part of a large anonymization procedure for edf
    """
    with Path(filename).open('r+b') as f:
        f.seek(168)
        f.write(16 * b' ')



from numpy import isnan, dtype
from numpy.testing import assert_almost_equal
from pytest import raises

from wonambi import Dataset

from .paths import openephys_dir


def test_openephys_dataset_01():
    d = Dataset(openephys_dir)
    data = d.read_data(begtime=1, endtime=2)

    d = Dataset(openephys_dir)
    data = d.read_data(chan=['CH1', ], begsam=10, endsam=1400)
    assert data.data[0][0, 0]  == -132.6

    mrk = d.read_markers()
    assert len(mrk) == 0



from os import SEEK_CUR, SEEK_SET, SEEK_END
from re import search, finditer, match
from datetime import datetime

from numpy import (fromfile,
                   frombuffer,
                   asmatrix,
                   array,
                   arange,
                   c_,
                   diff,
                   empty,
                   hstack,
                   ndarray,
                   NaN,
                   vstack,
                   where,
                   dtype,
                   float64,
                   int32,
                   uint8,
                   )

STATEVECTOR = ['Name', 'Length', 'Value', 'ByteLocation', 'BitLocation']


class BCI2000:
    """Basic class to read the data.

    Parameters
    ----------
    filename : path to file
        the name of the filename or directory
    """
    def __init__(self, filename):
        self.filename = filename

    def return_hdr(self):
        """Return the header for further use.

        Returns
        -------
        subj_id : str
            subject identification code
        start_time : datetime
            start time of the dataset
        s_freq : float
            sampling frequency
        chan_name : list of str
            list of all the channels
        n_samples : int
            number of samples in the dataset
        orig : dict
            additional information taken directly from the header

        Notes
        -----
        As far as I can, BCI2000 doesn't have channel labels, so we use dummies
        starting at chan001 (more consistent with Matlab 1-base indexing...)
        """
        orig = {}
        orig = _read_header(self.filename)

        nchan = int(orig['SourceCh'])
        chan_name = ['ch{:03d}'.format(i + 1) for i in range(nchan)]
        chan_dtype = dtype(orig['DataFormat'])
        self.statevector_len = int(orig['StatevectorLen'])

        s_freq = orig['Parameter']['SamplingRate']
        if s_freq.endswith('Hz'):
            s_freq = s_freq.replace('Hz', '')
        s_freq = int(s_freq.strip())
        self.s_freq = s_freq

        storagetime = orig['Parameter']['StorageTime'].replace('%20', ' ')
        try:  # newer version
            start_time = datetime.strptime(storagetime, '%a %b %d %H:%M:%S %Y')
        except:
            start_time = datetime.strptime(storagetime, '%Y-%m-%dT%H:%M:%S')

        subj_id = orig['Parameter']['SubjectName']

        self.dtype = dtype([(chan, chan_dtype) for chan in chan_name] +
                           [('statevector', 'S', self.statevector_len)])

        # compute n_samples based on file size - header
        with open(self.filename, 'rb') as f:
            f.seek(0, SEEK_END)
            EOData = f.tell()
        n_samples = int((EOData - int(orig['HeaderLen'])) / self.dtype.itemsize)

        self.s_freq = s_freq
        self.header_len = int(orig['HeaderLen'])
        self.n_samples = n_samples
        self.statevectors = _prepare_statevectors(orig['StateVector'])
        # TODO: a better way to parse header
        self.gain = array([float(x) for x in orig['Parameter']['SourceChGain'].split(' ')[1:]])

        return subj_id, start_time, s_freq, chan_name, n_samples, orig

    def return_dat(self, chan, begsam, endsam):
        """Return the data as 2D numpy.ndarray.

        Parameters
        ----------
        chan : int or list
            index (indices) of the channels to read
        begsam : int
            index of the first sample
        endsam : int
            index of the last sample

        Returns
        -------
        numpy.ndarray
            A 2d matrix, with dimension chan X samples
        """
        dat_begsam = max(begsam, 0)
        dat_endsam = min(endsam, self.n_samples)
        dur = dat_endsam - dat_begsam

        dtype_onlychan = dtype({k: v for k, v in self.dtype.fields.items() if v[0].kind != 'S'})

        # make sure we read some data at least, otherwise segfault
        if dat_begsam < self.n_samples and dat_endsam > 0:

            with self.filename.open('rb') as f:
                f.seek(self.header_len, SEEK_SET)  # skip header

                f.seek(self.dtype.itemsize * dat_begsam, SEEK_CUR)
                dat = fromfile(f, dtype=self.dtype, count=dur)

            dat = ndarray(dat.shape, dtype_onlychan, dat, 0, dat.strides).view((dtype_onlychan[0], len(dtype_onlychan.names))).T

        else:
            n_chan = len(dtype_onlychan.names)
            dat = empty((n_chan, 0))

        if begsam < 0:

            pad = empty((dat.shape[0], 0 - begsam))
            pad.fill(NaN)
            dat = c_[pad, dat]

        if endsam >= self.n_samples:

            pad = empty((dat.shape[0], endsam - self.n_samples))
            pad.fill(NaN)
            dat = c_[dat, pad]

        return dat[chan, :] * self.gain[chan][:, None]  # apply gain

    def return_markers(self, state='MicromedCode'):
        """Return all the markers (also called triggers or events).

        Returns
        -------
        list of dict
            where each dict contains 'name' as str, 'start' and 'end' as float
            in seconds from the start of the recordings, and 'chan' as list of
            str with the channels involved (if not of relevance, it's None).

        Raises
        ------
        FileNotFoundError
            when it cannot read the events for some reason (don't use other
            exceptions).
        """
        markers = []
        try:
            all_states = self._read_states()
        except ValueError:  # cryptic error when reading states
            return markers

        try:
            x = all_states[state]
        except KeyError:
            return markers

        markers = []
        i_mrk = hstack((0, where(diff(x))[0] + 1, len(x)))
        for i0, i1 in zip(i_mrk[:-1], i_mrk[1:]):
            marker = {'name': str(x[i0]),
                      'start': (i0) / self.s_freq,
                      'end': i1 / self.s_freq,
                     }
            markers.append(marker)

        return markers

    def _read_states(self):

        all_states = []
        with self.filename.open('rb') as f:
            f.seek(self.header_len, SEEK_SET)  # skip header
            StatevectorOffset = self.dtype.itemsize - self.statevector_len

            for i in range(self.n_samples):
                f.seek(StatevectorOffset, SEEK_CUR)
                raw_statevector = f.read(self.statevector_len)
                all_states.append(frombuffer(raw_statevector, dtype='<u1'))

        all_states = vstack(all_states).T

        states = {}
        for statename, statedef in self.statevectors.items():
            states[statename] = array(statedef['mult'] * asmatrix(all_states[statedef['slice'], :] & statedef['mask']), dtype=int32).squeeze()

        return states


def _read_header(filename):
    """It's a pain to parse the header. It might be better to use the cpp code
    but I would need to include it here.
    """
    header = _read_header_text(filename)
    first_row = header[0]
    EXTRA_ROWS = 3  # drop DefaultValue1 LowRange1 HighRange1

    hdr = {}
    for group in finditer('(\w*)= ([\w.]*)', first_row):
        hdr[group.group(1)] = group.group(2)

    if first_row.startswith('BCI2000V'):
        VERSION = hdr['BCI2000V']

    else:
        VERSION = '1'
        hdr['DataFormat'] = 'int16'

    for row in header[1:]:
        if row.startswith('['):  # remove '[ ... Definition ]'
            section = row[2:-14].replace(' ', '')

            if section == 'StateVector':
                hdr[section] = []
            else:
                hdr[section] = {} # defaultdict(dict)
            continue

        if row == '':
            continue

        elif section == 'StateVector':
            statevector = {key: value for key, value in list(zip(STATEVECTOR, row.split(' ')))}
            hdr[section].append(statevector)

        else:
            group = match('(?P<subsection>[\w:%]*) (?P<format>\w*) (?P<key>\w*)= (?P<value>.*) // ', row)
            onerow = group.groupdict()

            values = onerow['value'].split(' ')
            if len(values) > EXTRA_ROWS:
                value = ' '.join(onerow['value'].split(' ')[:-EXTRA_ROWS])
            else:
                value = ' '.join(values)

            hdr[section][onerow['key']] = value  # similar to matlab's output

    return hdr


def _read_header_length(filename):
    with filename.open('rb') as f:
        firstchar = f.read(100)  # should be enough to read the HeaderLen
        found = search('HeaderLen= (\d*) ', firstchar.decode())
        HeaderLen = int(found.group(1))

    return HeaderLen


def _read_header_text(filename):
    HeaderLen = _read_header_length(filename)
    with filename.open('rb') as f:
        header = f.read(HeaderLen).decode().split('\r\n')

    return header


def _prepare_statevectors(sv):

    statedefs = {}

    for v in sv:
        startbyte = int(v['ByteLocation'])
        startbit  = int(v['BitLocation'])
        nbits     = int(v['Length'])
        nbytes    = (startbit + nbits) // 8
        if (startbit + nbits) % 8:
            nbytes += 1
        extrabits = int(nbytes * 8) - nbits - startbit;
        startmask = 255 & (255 << startbit)
        endmask   = 255 & (255 >> extrabits)
        div       = (1 << startbit);
        v['slice'] = slice(startbyte, startbyte + nbytes)
        v['mask'] = array([255] * nbytes, dtype=uint8)
        v['mask'][0]  &= startmask
        v['mask'][-1] &= endmask
        v['mask'].shape = (nbytes, 1)
        v['mult'] = asmatrix(256.0 ** arange(nbytes, dtype=float64) / float(div))
        statedefs[v['Name']] = v

    return statedefs



from datetime import timedelta
from operator import itemgetter

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import (QAbstractItemView,
                             QComboBox,
                             QDateTimeEdit,
                             QDialog,
                             QDialogButtonBox,
                             QFileDialog,
                             QFormLayout,
                             QLabel,
                             QListWidget,
                             QListWidgetItem,
                             QPushButton,
                             QSpinBox,
                             )

from .utils import FormMenu, STAGE_NAME


class ChannelDialog(QDialog):
    """Template dialog for event detection.

    Attributes
    ----------
    parent : instance of QMainWindow
        the main window
    groups : list of dict
        information about groups from Channels
    index : dict of FormFloat
        Contains detection parameters.

    bbox : QDialogButtonBox
        Button box with Ok and Cancel
    idx_group : FormMenu
        Combo box of channel groups.
    idx_chan : QListWidget
        List widget containing all channels for selected group.
    idx_stage :  QListWidget
        List widget containing all stages.
    idx_cycle : QListWidget
        List widget containing all marked cycles.
    one_grp : dict
        info about selected group from Channels
    """
    def __init__(self, parent):
        super().__init__(None, Qt.WindowSystemMenuHint | Qt.WindowTitleHint)
        self.parent = parent

        self.setWindowModality(Qt.WindowModal)
        self.groups = self.parent.channels.groups
        self.index = {}
        self.cycles = None

        self.create_widgets()

    def create_widgets(self):
        """Build basic components of dialog."""
        self.bbox = QDialogButtonBox(
                QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        self.idx_ok = self.bbox.button(QDialogButtonBox.Ok)
        self.idx_cancel = self.bbox.button(QDialogButtonBox.Cancel)

        self.idx_group = FormMenu([gr['name'] for gr in self.groups])

        chan_box = QListWidget()
        self.idx_chan = chan_box

        stage_box = QListWidget()
        stage_box.addItems(STAGE_NAME)
        stage_box.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self.idx_stage = stage_box

        cycle_box = QListWidget()
        cycle_box.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self.idx_cycle = cycle_box

    def update_groups(self):
        """Update the channel groups list when dialog is opened."""
        self.groups = self.parent.channels.groups
        self.idx_group.clear()
        for gr in self.groups:
            self.idx_group.addItem(gr['name'])

        self.update_channels()

    def update_channels(self):
        """Update the channels list when a new group is selected."""
        group_dict = {k['name']: i for i, k in enumerate(self.groups)}
        group_index = group_dict[self.idx_group.currentText()]
        self.one_grp = self.groups[group_index]

        self.idx_chan.clear()

        self.idx_chan.setSelectionMode(QAbstractItemView.ExtendedSelection)
        for chan in self.one_grp['chan_to_plot']:
            name = chan + '—(' + '+'.join(self.one_grp['ref_chan']) + ')'
            item = QListWidgetItem(name)
            self.idx_chan.addItem(item)

    def update_cycles(self):
        """Enable cycles checkbox only if there are cycles marked, with no
        errors."""
        self.idx_cycle.clear()

        try:
            self.cycles = self.parent.notes.annot.get_cycles()

        except ValueError as err:
            self.idx_cycle.setEnabled(False)
            msg = 'There is a problem with the cycle markers: ' + str(err)
            self.parent.statusBar().showMessage(msg)

        else:
            if self.cycles is None:
                self.idx_cycle.setEnabled(False)
            else:
                self.idx_cycle.setEnabled(True)
                for i in range(len(self.cycles)):
                    self.idx_cycle.addItem(str(i+1))

    def get_channels(self):
        """Get the selected channel(s in order).

        Returns
        -------
        list of str
            name of each channel (without group), in original record order
        """
        selectedItems = self.idx_chan.selectedItems()
        selected_chan = [x.text().split('—')[0] for x in selectedItems]
        chan_in_order = []
        for chan in self.one_grp['chan_to_plot']:
            if chan in selected_chan:
                chan_in_order.append(chan)

        return chan_in_order

    def get_cycles(self):
        """Get the selected cycle(s in order).

        Returns
        -------
        list of tuple
            Each tuple is (start time (sec), end time (sec), index (starting
            at 1)."""
        idx_cyc_sel = [
                int(x.text()) - 1 for x in self.idx_cycle.selectedItems()]
        if not idx_cyc_sel:
            cycle = None
        else:
            cycle = itemgetter(*idx_cyc_sel)(self.cycles)
            if len(idx_cyc_sel) == 1:
                cycle = [cycle]

        return cycle


class DateTimeDialog(QDialog):
    """Dialog to specify time in the recordings, either as seconds from the
    start of the recordings or absolute time.

    Parameters
    ----------
    title : str
        'Lights out' or 'Lights on'
    start_time : datetime
        absolute start time of the recordings
    dur : int
        total duration of the recordings

    Notes
    -----
    The value of interest is in self.idx_seconds.value(), which is seconds
    from the start of the recordings.
    """
    def __init__(self, title, start_time, dur):
        super().__init__()

        self.start_time = start_time
        self.dur = dur
        end_time = start_time + timedelta(seconds=dur)

        self.setWindowTitle(title)

        bbox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        self.idx_ok = bbox.button(QDialogButtonBox.Ok)
        self.idx_cancel = bbox.button(QDialogButtonBox.Cancel)

        bbox.clicked.connect(self.button_clicked)

        self.idx_seconds = QSpinBox()
        self.idx_seconds.setMinimum(0)
        self.idx_seconds.setMaximum(dur)
        self.idx_seconds.valueChanged.connect(self.changed_spin)

        self.idx_datetime = QDateTimeEdit(start_time)
        self.idx_datetime.setMinimumDate(start_time)
        self.idx_datetime.setMaximumDate(end_time)
        self.idx_datetime.setDisplayFormat('dd-MMM-yyyy HH:mm:ss')
        self.idx_datetime.dateTimeChanged.connect(self.changed_datetime)

        layout = QFormLayout()
        layout.addRow('', QLabel('Enter ' + title + ' time'))
        layout.addRow('Seconds from recording start', self.idx_seconds)
        layout.addRow('Absolute time', self.idx_datetime)
        layout.addRow(bbox)

        self.setLayout(layout)

    def button_clicked(self, button):
        if button == self.idx_ok:
            self.accept()

        elif button == self.idx_cancel:
            self.reject()

    def changed_spin(self, i):
        self.idx_datetime.blockSignals(True)
        self.idx_datetime.setDateTime(self.start_time + timedelta(seconds=i))
        self.idx_datetime.blockSignals(False)

    def changed_datetime(self, dt):
        val = (dt.toPyDateTime() - self.start_time).total_seconds()

        if val < 0 or val >= self.dur:
            val = min(self.dur, max(val, 0))
            self.changed_spin(val)

        self.idx_seconds.blockSignals(True)
        self.idx_seconds.setValue(val)
        self.idx_seconds.blockSignals(False)


class SVGDialog(QDialog):

    def __init__(self, dirname):
        super().__init__()

        self.dirname = str(dirname)

        self.setWindowTitle('Export to SVG')

        bbox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        self.idx_ok = bbox.button(QDialogButtonBox.Ok)
        self.idx_cancel = bbox.button(QDialogButtonBox.Cancel)
        bbox.clicked.connect(self.button_clicked)

        self.idx_list = QComboBox()
        self.idx_list.addItems(['Traces', 'Overview'])

        self.idx_file = QPushButton()
        self.idx_file.setText('(click to choose file)')
        self.idx_file.clicked.connect(self.select_filename)

        layout = QFormLayout()
        layout.addRow('Which Panel', self.idx_list)
        layout.addRow('File Name', self.idx_file)
        layout.addRow(bbox)

        self.setLayout(layout)

    def button_clicked(self, button):
        if button == self.idx_ok:
            self.accept()

        elif button == self.idx_cancel:
            self.reject()

    def select_filename(self):
        filename, _ = QFileDialog.getSaveFileName(
            self, 'Export screenshot', self.dirname, 'Image (*.svg)')

        if filename == '':
            return

        self.idx_file.setText(filename)



from datetime import datetime
from PyQt5.QtCore import QDateTime

from wonambi.widgets.modal_widgets import DateTimeDialog
from .paths import GUI_PATH


start_time = datetime(2018, 1, 1, 12, 0, 0)
duration = 120


def test_widget_labels(qtbot):
    d = DateTimeDialog('lights out', start_time, duration)
    qtbot.addWidget(d)

    d.idx_seconds.setValue(-10)
    assert d.idx_seconds.value() == 0

    d.idx_seconds.setValue(duration + 10)
    assert d.idx_seconds.value() == duration

    new_datetime = QDateTime(2018, 1, 1, 11, 30, 0)
    d.changed_datetime(new_datetime)
    assert d.idx_seconds.value() == 0

    # how to test this explicitly?
    d.button_clicked(d.idx_ok)
    d.button_clicked(d.idx_cancel)

    d.grab().save(str(GUI_PATH / 'analysis_02_timedialog.png'))



"""Widget to show power spectrum.
"""
from logging import getLogger

from numpy import log, ceil, floor, min
from scipy.signal import welch
from PyQt5.QtCore import Qt
from PyQt5.QtGui import (QColor,
                         QPen,
                         )
from PyQt5.QtWidgets import (QComboBox,
                             QFormLayout,
                             QGraphicsView,
                             QGraphicsScene,
                             QGroupBox,
                             QVBoxLayout,
                             QWidget,
                             )

from .utils import Path, LINE_WIDTH, LINE_COLOR, FormFloat, FormBool
from .settings import Config

lg = getLogger(__name__)


class ConfigSpectrum(Config):

    def __init__(self, update_widget):
        super().__init__('spectrum', update_widget)

    def create_config(self):

        box0 = QGroupBox('Spectrum')

        for k in self.value:
            self.index[k] = FormFloat()
        self.index['log'] = FormBool('Log-transform')

        form_layout = QFormLayout()
        form_layout.addRow('Min X', self.index['x_min'])
        form_layout.addRow('Max X', self.index['x_max'])
        form_layout.addRow('Ticks on X-axis', self.index['x_tick'])
        form_layout.addRow('Min Y', self.index['y_min'])
        form_layout.addRow('Max Y', self.index['y_max'])
        form_layout.addRow('Ticks on Y-axis', self.index['y_tick'])

        form_layout.addWidget(self.index['log'])
        box0.setLayout(form_layout)

        main_layout = QVBoxLayout()
        main_layout.addWidget(box0)
        main_layout.addStretch(1)

        self.setLayout(main_layout)


class Spectrum(QWidget):
    """Plot the power spectrum for a specified channel.

    Attributes
    ----------
    parent : instance of QMainWindow
        the main window.
    x_limit : tuple or list
        2 values specifying the limit on x-axis
    y_limit : tuple or list
        2 values specifying the limit on y-axis
    log : bool
        log-transform the data or not
    idx_chan : instance of QComboBox
        the element with the list of channel names.
    idx_x_min : instance of QLineEdit
        value with min x value
    idx_x_max : instance of QLineEdit
        value with max x value
    idx_y_min : instance of QLineEdit
        value with min y value
    idx_y_max : instance of QLineEdit
        value with max y value
    idx_log : instance of QCheckBox
        widget that defines if log should be used or not
    idx_fig : instance of QGraphicsView
        the view with the power spectrum
    scene : instance of QGraphicsScene
        the scene with GraphicsItems

    Notes
    -----
    If data contains NaN, it doesn't create any spectrum (feature or bug?).
    """
    def __init__(self, parent):
        super().__init__()
        self.parent = parent

        self.config = ConfigSpectrum(self.display_window)

        self.selected_chan = None
        self.idx_chan = None
        self.idx_fig = None
        self.scene = None

        self.create()

    def create(self):
        """Create empty scene for power spectrum."""
        self.idx_chan = QComboBox()
        self.idx_chan.activated.connect(self.display_window)

        self.idx_fig = QGraphicsView(self)
        self.idx_fig.scale(1, -1)

        layout = QVBoxLayout()
        layout.addWidget(self.idx_chan)
        layout.addWidget(self.idx_fig)
        self.setLayout(layout)

        self.resizeEvent(None)

    def show_channame(self, chan_name):
        self.selected_chan = self.idx_chan.currentIndex()

        self.idx_chan.clear()
        self.idx_chan.addItem(chan_name)
        self.idx_chan.setCurrentIndex(0)

    def update(self):
        """Add channel names to the combobox."""
        self.idx_chan.clear()
        for chan_name in self.parent.traces.chan:
            self.idx_chan.addItem(chan_name)

        if self.selected_chan is not None:
            self.idx_chan.setCurrentIndex(self.selected_chan)
            self.selected_chan = None

    def display_window(self):
        """Read the channel name from QComboBox and plot its spectrum.

        This function is necessary it reads the data and it sends it to
        self.display. When the user selects a smaller chunk of data from the
        visible traces, then we don't need to call this function.
        """
        if self.idx_chan.count() == 0:
            self.update()

        chan_name = self.idx_chan.currentText()
        lg.debug('Power spectrum for channel ' + chan_name)

        if chan_name:
            trial = 0
            data = self.parent.traces.data(trial=trial, chan=chan_name)
            self.display(data)
        else:
            self.scene.clear()

    def display(self, data):
        """Make graphicsitem for spectrum figure.

        Parameters
        ----------
        data : ndarray
            1D vector containing the data only

        This function can be called by self.display_window (which reads the
        data for the selected channel) or by the mouse-events functions in
        traces (which read chunks of data from the user-made selection).
        """
        value = self.config.value
        self.scene = QGraphicsScene(value['x_min'], value['y_min'],
                                    value['x_max'] - value['x_min'],
                                    value['y_max'] - value['y_min'])
        self.idx_fig.setScene(self.scene)

        self.add_grid()
        self.resizeEvent(None)

        s_freq = self.parent.traces.data.s_freq
        f, Pxx = welch(data, fs=s_freq,
                       nperseg=int(min((s_freq, len(data)))))  # force int

        freq_limit = (value['x_min'] <= f) & (f <= value['x_max'])

        if self.config.value['log']:
            Pxx_to_plot = log(Pxx[freq_limit])
        else:
            Pxx_to_plot = Pxx[freq_limit]

        self.scene.addPath(Path(f[freq_limit], Pxx_to_plot),
                           QPen(QColor(LINE_COLOR), LINE_WIDTH))

    def add_grid(self):
        """Add axis and ticks to figure.

        Notes
        -----
        I know that visvis and pyqtgraphs can do this in much simpler way, but
        those packages create too large a padding around the figure and this is
        pretty fast.

        """
        value = self.config.value

        # X-AXIS
        # x-bottom
        self.scene.addLine(value['x_min'], value['y_min'],
                           value['x_min'], value['y_max'],
                           QPen(QColor(LINE_COLOR), LINE_WIDTH))
        # at y = 0, dashed
        self.scene.addLine(value['x_min'], 0,
                           value['x_max'], 0,
                           QPen(QColor(LINE_COLOR), LINE_WIDTH, Qt.DashLine))
        # ticks on y-axis
        y_high = int(floor(value['y_max']))
        y_low = int(ceil(value['y_min']))
        x_length = (value['x_max'] - value['x_min']) / value['x_tick']
        for y in range(y_low, y_high):
            self.scene.addLine(value['x_min'], y,
                               value['x_min'] + x_length, y,
                               QPen(QColor(LINE_COLOR), LINE_WIDTH))
        # Y-AXIS
        # left axis
        self.scene.addLine(value['x_min'], value['y_min'],
                           value['x_max'], value['y_min'],
                           QPen(QColor(LINE_COLOR), LINE_WIDTH))
        # larger ticks on x-axis every 10 Hz
        x_high = int(floor(value['x_max']))
        x_low = int(ceil(value['x_min']))
        y_length = (value['y_max'] - value['y_min']) / value['y_tick']
        for x in range(x_low, x_high, 10):
            self.scene.addLine(x, value['y_min'],
                               x, value['y_min'] + y_length,
                               QPen(QColor(LINE_COLOR), LINE_WIDTH))
        # smaller ticks on x-axis every 10 Hz
        y_length = (value['y_max'] - value['y_min']) / value['y_tick'] / 2
        for x in range(x_low, x_high, 5):
            self.scene.addLine(x, value['y_min'],
                               x, value['y_min'] + y_length,
                               QPen(QColor(LINE_COLOR), LINE_WIDTH))

    def resizeEvent(self, event):
        """Fit the whole scene in view.

        Parameters
        ----------
        event : instance of Qt.Event
            not important

        """
        value = self.config.value
        self.idx_fig.fitInView(value['x_min'],
                               value['y_min'],
                               value['x_max'] - value['x_min'],
                               value['y_max'] - value['y_min'])

    def reset(self):
        """Reset widget as new"""
        self.idx_chan.clear()
        if self.scene is not None:
            self.scene.clear()
        self.scene = None



"""Analysis and export convenience functions.
"""

from logging import getLogger
from itertools import compress
from csv import writer
from numpy import (amax, amin, asarray, concatenate, in1d, mean, negative, ptp, 
                   reshape, sqrt, square)

try:
    from PyQt5.QtCore import Qt
    from PyQt5.QtWidgets import QProgressDialog
except ImportError:
    Qt = None
    QProgressDialog = None

from .. import __version__
from .math import math, get_descriptives
from .frequency import band_power
from .peaks import get_slopes

lg = getLogger(__name__)


def event_params(segments, params, band=None, n_fft=None, slopes=None, 
                 prep=None, parent=None):
    """Compute event parameters.
    
    Parameters
    ----------
    segments : instance of wonambi.trans.select.Segments
        list of segments, with time series and metadata
    params : dict of bool, or str
        'dur', 'minamp', 'maxamp', 'ptp', 'rms', 'power', 'peakf', 'energy', 
        'peakef'. If 'all', a dict will be created with these keys and all 
        values as True, so that all parameters are returned.
    band : tuple of float
        band of interest for power and energy
    n_fft : int
        length of FFT. if shorter than input signal, signal is truncated; if 
        longer, signal is zero-padded to length
    slopes : dict of bool
        'avg_slope', 'max_slope', 'prep', 'invert'
    prep : dict of bool
        same keys as params. if True, segment['trans_data'] will be used as dat
    parent : QMainWindow
        for use with GUI only
        
    Returns
    -------
    list of dict
        list of segments, with time series, metadata and parameters
    """
    if parent is not None:
        progress = QProgressDialog('Computing parameters', 'Abort',
                                   0, len(segments) - 1, parent)
        progress.setWindowModality(Qt.ApplicationModal)

    param_keys = ['dur', 'minamp', 'maxamp', 'ptp', 'rms', 'power', 'peakpf', 
                  'energy', 'peakef']
    
    if params == 'all':
        params = {k: 1 for k in param_keys}
    if prep is None:
        prep = {k: 0 for k in param_keys}    
    if band is None:
        band = (None, None)

    params_out = []
    evt_output = False

    for i, seg in enumerate(segments):
        out = dict(seg)
        dat = seg['data']            

        if params['dur']:
            out['dur'] = float(dat.number_of('time')) / dat.s_freq
            evt_output = True

        if params['minamp']:
            dat1 = dat
            if prep['minamp']:
                dat1 = seg['trans_data']
            out['minamp'] = math(dat1, operator=_amin, axis='time')
            evt_output = True

        if params['maxamp']:
            dat1 = dat
            if prep['maxamp']:
                dat1 = seg['trans_data']
            out['maxamp'] = math(dat1, operator=_amax, axis='time')
            evt_output = True

        if params['ptp']:
            dat1 = dat
            if prep['ptp']:
                dat1 = seg['trans_data']
            out['ptp'] = math(dat1, operator=_ptp, axis='time')
            evt_output = True

        if params['rms']:
            dat1 = dat
            if prep['rms']:
                dat1 = seg['trans_data']
            out['rms'] = math(dat1, operator=(square, _mean, sqrt),
               axis='time')
            evt_output = True

        for pw, pk in [('power', 'peakpf'), ('energy', 'peakef')]:

            if params[pw] or params[pk]:
                evt_output = True

                if prep[pw] or prep[pk]:
                    prep_pw, prep_pk = band_power(seg['trans_data'], band,
                                                 scaling=pw, n_fft=n_fft)
                if not (prep[pw] and prep[pk]):
                    raw_pw, raw_pk = band_power(dat, band, 
                                                scaling=pw, n_fft=n_fft)

                if prep[pw]:
                    out[pw] = prep_pw
                else:
                    out[pw] = raw_pw

                if prep[pk]:
                    out[pk] = prep_pk
                else:
                    out[pk] = raw_pk

        if slopes:
            evt_output = True
            out['slope'] = {}
            dat1 = dat
            if slopes['prep']:
                dat1 = seg['trans_data']
            if slopes['invert']:
                dat1 = math(dat1, operator=negative, axis='time')

            if slopes['avg_slope'] and slopes['max_slope']:
                level = 'all'
            elif slopes['avg_slope']:
                level = 'average'
            else:
                level = 'maximum'

            for chan in dat1.axis['chan'][0]:
                d = dat1(chan=chan)[0]
                out['slope'][chan] = get_slopes(d, dat.s_freq, level=level)
                
        if evt_output:
            timeline = dat.axis['time'][0]
            out['start'] = timeline[0]
            out['end'] = timeline[-1]
            params_out.append(out)

        if parent:
            progress.setValue(i)
            if progress.wasCanceled():
                msg = 'Analysis canceled by user.'
                parent.statusBar().showMessage(msg)
                return

    if parent:
        progress.close()

    return params_out

def export_event_params(filename, params, count=None, density=None):
    """Write event analysis data to CSV."""
    heading_row_1 = ['Segment index',
                   'Start time',
                   'End time',
                   'Stitches',
                   'Stage',
                   'Cycle',
                   'Event type',
                   'Channel']
    spacer = [''] * (len(heading_row_1) - 1)

    param_headings_1 = ['Min. amplitude (uV)',
                        'Max. amplitude (uV)',
                        'Peak-to-peak amplitude (uV)',
                        'RMS (uV)']
    param_headings_2 = ['Power (uV**2)',
                        'Peak power frequency (Hz)',
                        'Energy (uV**2/s)',
                        'Peak energy frequency (Hz)']
    slope_headings =   ['Q1 average slope (uV/s)',
                        'Q2 average slope (uV/s)',
                        'Q3 average slope (uV/s)',
                        'Q4 average slope (uV/s)',
                        'Q23 average slope (uV/s)',
                        'Q1 max. slope (uV/s**2)',
                        'Q2 max. slope (uV/s**2)',
                        'Q3 max. slope (uV/s**2)',
                        'Q4 max. slope (uV/s**2)',
                        'Q23 max. slope (uV/s**2)']
    ordered_params_1 = ['minamp', 'maxamp', 'ptp', 'rms']
    ordered_params_2 = ['power', 'peakpf', 'energy', 'peakef']

    idx_params_1 = in1d(ordered_params_1, list(params[0].keys()))
    sel_params_1 = list(compress(ordered_params_1, idx_params_1))
    heading_row_2 = list(compress(param_headings_1, idx_params_1))

    if 'dur' in params[0].keys():
        heading_row_2 = ['Duration (s)'] + heading_row_2

    idx_params_2 = in1d(ordered_params_2, list(params[0].keys()))
    sel_params_2 = list(compress(ordered_params_2, idx_params_2))
    heading_row_3 = list(compress(param_headings_2, idx_params_2))

    heading_row_4 = []
    if 'slope' in params[0].keys():
        if next(iter(params[0]['slope']))[0]:
            heading_row_4.extend(slope_headings[:5])
        if next(iter(params[0]['slope']))[1]:
            heading_row_4.extend(slope_headings[5:])

    # Get data as matrix and compute descriptives
    dat = []
    if 'dur' in params[0].keys():
        one_mat = asarray([seg['dur'] for seg in params \
                           for chan in seg['data'].axis['chan'][0]])
        one_mat = reshape(one_mat, (len(one_mat), 1))
        dat.append(one_mat)

    if sel_params_1:
        one_mat = asarray([[seg[x](chan=chan)[0] for x in sel_params_1] \
                for seg in params for chan in seg['data'].axis['chan'][0]])
        dat.append(one_mat)

    if sel_params_2:
        one_mat = asarray([[seg[x][chan] for x in sel_params_2] \
                for seg in params for chan in seg['data'].axis['chan'][0]])
        dat.append(one_mat)

    if 'slope' in params[0].keys():
        one_mat = asarray([[x for y in seg['slope'][chan] for x in y] \
                for seg in params for chan in seg['data'].axis['chan'][0]])
        dat.append(one_mat)

    if dat:
        dat = concatenate(dat, axis=1)
        desc = get_descriptives(dat)

    with open(filename, 'w', newline='') as f:
        lg.info('Writing to ' + str(filename))
        csv_file = writer(f)
        csv_file.writerow(['Wonambi v{}'.format(__version__)])

        if count:
            csv_file.writerow(['Count', count])
        if density:
            csv_file.writerow(['Density', density])

        if dat == []:
            return

        csv_file.writerow(heading_row_1 + heading_row_2 + heading_row_3 \
                          + heading_row_4)
        csv_file.writerow(['Mean'] + spacer + list(desc['mean']))
        csv_file.writerow(['SD'] + spacer + list(desc['sd']))
        csv_file.writerow(['Mean of ln'] + spacer + list(desc['mean_log']))
        csv_file.writerow(['SD of ln'] + spacer + list(desc['sd_log']))
        idx = 0

        for seg in params:
            if seg['cycle'] is not None:
                seg['cycle'] = seg['cycle'][2]

            for chan in seg['data'].axis['chan'][0]:
                idx += 1
                data_row_1 = [seg[x](chan=chan)[0] for x in sel_params_1]
                data_row_2 = [seg[x][chan] for x in sel_params_2]

                if 'dur' in seg.keys():
                    data_row_1 = [seg['dur']] + data_row_1

                if 'slope' in seg.keys():
                    data_row_3 = [x for y in seg['slope'][chan] for x in y]
                    data_row_2 = data_row_2 + data_row_3

                csv_file.writerow([idx,
                                   seg['start'],
                                   seg['end'],
                                   seg['n_stitch'],
                                   seg['stage'],
                                   seg['cycle'],
                                   seg['name'],
                                   chan,
                                   ] + data_row_1 + data_row_2)

def export_freq(xfreq, filename, desc=None):
    """Write frequency analysis data to CSV.

    Parameters
    ----------
    xfreq : list of dict
        spectral data, one dict per segment, where 'data' is ChanFreq
    filename : str
        output filename
    desc : dict of ndarray
        descriptives
    '"""
    heading_row_1 = ['Segment index',
                   'Start time',
                   'End time',
                   'Duration',
                   'Stitches',
                   'Stage',
                   'Cycle',
                   'Event type',
                   'Channel',
                   ]
    spacer = [''] * (len(heading_row_1) - 1)
    freq = list(xfreq[0]['data'].axis['freq'][0])

    with open(filename, 'w', newline='') as f:
        lg.info('Writing to ' + str(filename))
        csv_file = writer(f)
        csv_file.writerow(['Wonambi v{}'.format(__version__)])
        csv_file.writerow(heading_row_1 + freq)

        if desc:
            csv_file.writerow(['Mean'] + spacer + list(desc['mean']))
            csv_file.writerow(['SD'] + spacer + list(desc['sd']))
            csv_file.writerow(['Mean of ln'] + spacer + list(
                    desc['mean_log']))
            csv_file.writerow(['SD of ln'] + spacer + list(desc['sd_log']))

        idx = 0
        for seg in xfreq:

            for chan in seg['data'].axis['chan'][0]:
                idx += 1

                cyc = None
                if seg['cycle'] is not None:
                    cyc = seg['cycle'][2]

                data_row = list(seg['data'](chan=chan)[0])
                csv_file.writerow([idx,
                                   seg['start'],
                                   seg['end'],
                                   seg['duration'],
                                   seg['n_stitch'],
                                   seg['stage'],
                                   cyc,
                                   seg['name'],
                                   chan,
                                   ] + data_row)


def export_freq_band(xfreq, bands, filename):
    """Write frequency analysis data to CSV by pre-defined band."""
    heading_row_1 = ['Segment index',
                   'Start time',
                   'End time',
                   'Duration',
                   'Stitches',
                   'Stage',
                   'Cycle',
                   'Event type',
                   'Channel',
                   ]
    spacer = [''] * (len(heading_row_1) - 1)
    band_hdr = [str(b1) + '-' + str(b2) for b1, b2 in bands]
    xband = xfreq.copy()

    for seg in xband:
        bandlist = []

        for i, b in enumerate(bands):
            pwr, _ = band_power(seg['data'], b)
            bandlist.append(pwr)

        seg['band'] = bandlist

    as_matrix = asarray([
            [x['band'][y][chan] for y in range(len(x['band']))] \
            for x in xband for chan in x['band'][0].keys()])
    desc = get_descriptives(as_matrix)

    with open(filename, 'w', newline='') as f:
        lg.info('Writing to ' + str(filename))
        csv_file = writer(f)
        csv_file.writerow(['Wonambi v{}'.format(__version__)])
        csv_file.writerow(heading_row_1 + band_hdr)
        csv_file.writerow(['Mean'] + spacer + list(desc['mean']))
        csv_file.writerow(['SD'] + spacer + list(desc['sd']))
        csv_file.writerow(['Mean of ln'] + spacer + list(desc['mean_log']))
        csv_file.writerow(['SD of ln'] + spacer + list(desc['sd_log']))
        idx = 0

        for seg in xband:

            for chan in seg['band'][0].keys():
                idx += 1

                cyc = None
                if seg['cycle'] is not None:
                    cyc = seg['cycle'][2]

                data_row = list(
                        [seg['band'][x][chan] for x in range(
                                len(seg['band']))])
                csv_file.writerow([idx,
                                   seg['start'],
                                   seg['end'],
                                   seg['duration'],
                                   seg['n_stitch'],
                                   seg['stage'],
                                   cyc,
                                   seg['name'],
                                   chan,
                                   ] + data_row)

def _amax(x, axis, keepdims=None):
    return amax(x, axis=axis)

def _amin(x, axis, keepdims=None):
    return amin(x, axis=axis)

def _ptp(x, axis, keepdims=None):
    return ptp(x, axis=axis)

def _mean(x, axis, keepdims=None):
    return mean(x, axis=axis)


"""This module contains classes of graphoelements.

These graphoelements can be generated by the package "detect".

"""
from copy import deepcopy
from csv import reader, writer
from numpy import empty, asarray, sum, mean

from .datatype import Data


class Graphoelement:
    """Class containing all the events of one type in one dataset.

    Attributes
    ----------
    chan_name : ndarray (dtype='U')
        list of channels
        
    """
    def __init__(self):
        self.chan_name = None
        self.events = []

    def __iter__(self):
        for one_event in self.events:
            yield one_event
            
    def __len__(self):
        return len(self.events)
    
    def __getitem__(self, index):
        return self.events[index]

    def __call__(self, func=None):

        events = []
        for one_ev in self.events:
            if func(one_ev):
                events.append(one_ev)

        output = deepcopy(self)
        output.events = events

        return output

    def to_data(self, parameter, operator=mean):

        data = Data()
        data.axis = {'chan': empty(1, dtype='O')}
        data.axis['chan'][0] = self.chan_name
        data.data = empty(1, dtype='O')

        values = []
        for one_chan in self.chan_name:
            if parameter == 'count':
                value = sum(1 for x in self.events if x['chan'] == one_chan)
            else:
                value = operator([x[parameter] for x in self.events
                                  if x['chan'] == one_chan])
            values.append(value)

        data.data[0] = asarray(values)
        return data
    
    def to_annot(self, annot, name):
        """Write events to Annotations file.
        
        Parameters
        ----------
        annot : instance of Annotations
            Annotations file
        name : str
            name for the event type        
        """
        for one_ev in self.events:
            annot.add_event(name,
                            (one_ev['start'], one_ev['end']),
                            chan=one_ev['chan'])


class Ripple(Graphoelement):
    def __init__(self):
        super().__init__()
        pass


class SlowWaves(Graphoelement):
    """Class containing all the slow waves in one dataset.

    Attributes
    ----------
    events : list of dict
        list of slow waves, where each SW contains:
            - start_time : float
                start time of the SW
            - trough_time : float
                time of the lowest value
            - zero_time : float
                time of the neg to pos zero-crossing
            - peak_time : float
                time of the highest value
            - end_time : float
                end time of the SW
            - trough_val : float
                the lowest value
            - peak_val : float
                the highest value
            - dur: float
                duration of the SW
            - area_under_curve : float
                sum of all values divided by duration
            - ptp : float
                peak-to-peak (difference between highest and lowest value)
            - chan': str
                channel label
    
    """
    def __init__(self):
        super().__init__()
        
        one_sw = {'start_time': None,
                  'trough_time': None,
                  'zero_time': None,
                  'peak_time': None,
                  'end_time': None,
                  'trough_val': None,
                  'peak_val': None,
                  'dur': None,
                  'area_under_curve': None,
                  'ptp': None,
                  'chan': [],
                  }
        self.events.append(one_sw)


class Spindles(Graphoelement):
    """Class containing all the spindles in one dataset.

    Attributes
    ----------
    mean : ndarray (dtype='float')
        mean of each channel
    std : ndarray (dtype='float')
        standard deviation of each channel
    det_value : ndarray (dtype='float')
        value used for detection for each channel
    sel_value : ndarray (dtype='float')
        value used for selection for each channel
    events : list of dict
        list of spindles, where each spindle contains:
            - start_time : float
                start time of the spindle
            - end_time : float
                end time of the spindle
            - peak_time : float
                time of the highest value
            - peak_val : float
                the highest value
            - chan': str
                channel label

    """
    def __init__(self):
        super().__init__()
        self.mean = None
        self.std = None
        self.det_value = None
        self.sel_value = None
        self.density = None

        one_spindle = {'start_time': None,
                       'end_time': None,
                       'chan': [],
                       'peak_val': None,
                       'peak_time': None,
                       }
        self.events.append(one_spindle)


def events_from_csv(source_file, stage=[]):
    """Import events from Wonambi event CSV.
    
    Parameters
    ----------
    source_file : str
        path to file CSV file
    stage : list of str (optional)
        target stage, eg ['NREM2', 'NREM3'] or ['REM']
        
    Returns
    -------
    instance of wonambi.Graphoelement
        class with events list
    """
    events = []    
    stage_cond = True
    
    with open(source_file, 'r', encoding='utf-8') as csvfile:
        csv_reader = reader(csvfile, delimiter=',')
        
        for row in csv_reader:
            try:
                int(row[0])
                if stage:
                    stage_cond = row[4] in stage
                if stage_cond:
                    one_ev = {'name': row[6],
                              'start': float(row[1]),
                              'end': float(row[2]),
                              'chan': row[7].split(', '),  # always a list
                              'stage': row[4],
                              'quality': 'Good'
                              }
                    events.append(one_ev)
                
            except ValueError:
                continue
    
    grapho = Graphoelement()
    grapho.events = events
    grapho.chan_name = list(set([chan for e in events for chan in e['chan']]))
    
    return grapho



"""Convenient module to convert data based on simple mathematical operations.
"""
from inspect import getfullargspec
from logging import getLogger

# for Math
from numpy import (absolute,
                   angle,
                   amax, 
                   amin,
                   asarray,
                   concatenate,
                   diff,
                   empty,
                   exp,
                   gradient,
                   in1d,
                   isinf,
                   log,
                   log10,
                   median,
                   mean,
                   nan,
                   nanmean,
                   nanstd,
                   negative,
                   ones,
                   pad,
                   ptp,
                   reshape,
                   sign,
                   sqrt,
                   square,
                   sum,
                   std,
                   where,
                   unwrap)
from scipy.signal import detrend, hilbert, fftconvolve
from scipy.stats import mode

lg = getLogger(__name__)

NOKEEPDIM = (median, mode)


def math(data, operator=None, operator_name=None, axis=None):
    """Apply mathematical operation to each trial and channel individually.

    Parameters
    ----------
    data : instance of DataTime, DataFreq, or DataTimeFreq

    operator : function or tuple of functions, optional
        function(s) to run on the data.
    operator_name : str or tuple of str, optional
        name of the function(s) to run on the data.
    axis : str, optional
        for functions that accept it, which axis you should run it on.

    Returns
    -------
    instance of Data
        data where the trials underwent operator.

    Raises
    ------
    TypeError
        If you pass both operator and operator_name.
    ValueError
        When you try to operate on an axis that has already been removed.

    Notes
    -----
    operator and operator_name are mutually exclusive. operator_name is given
    as shortcut for most common operations.

    If a function accepts an 'axis' argument, you need to pass 'axis' to the
    constructor. In this way, it'll apply the function to the correct
    dimension.

    The possible point-wise operator_name are:
    'absolute', 'angle', 'dB' (=10 * log10), 'exp', 'log', 'sqrt', 'square',
    'unwrap'

    The operator_name's that need an axis, but do not remove it:
    'hilbert', 'diff', 'detrend'

    The operator_name's that need an axis and remove it:
    'mean', 'median', 'mode', 'std'

    Examples
    --------
    You can pass a single value or a tuple. The order starts from left to
    right, so abs of the hilbert transform, should be:

    >>> rms = math(data, operator_name=('hilbert', 'abs'), axis='time')

    If you want to pass the power of three, use lambda (or partial):

    >>> p3 = lambda x: power(x, 3)
    >>> data_p3 = math(data, operator=p3)

    Note that lambdas are fine with point-wise operation, but if you want them
    to operate on axis, you need to pass ''axis'' as well, so that:

    >>> std_ddof = lambda x, axis: std(x, axis, ddof=1)
    >>> data_std = math(data, operator=std_ddof)

    If you don't pass 'axis' in lambda, it'll never know on which axis the
    function should be applied and you'll get unpredictable results.

    If you want to pass a function that operates on an axis and removes it (for
    example, if you want the max value over time), you need to add an argument
    in your function called ''keepdims'' (the values won't be used):

    >>> def func(x, axis, keepdims=None):
    >>>     return nanmax(x, axis=axis)
    """
    if operator is not None and operator_name is not None:
        raise TypeError('Parameters "operator" and "operator_name" are '
                        'mutually exclusive')

    # turn input into a tuple of functions in operators
    if operator_name is not None:
        if isinstance(operator_name, str):
            operator_name = (operator_name, )

        operators = []
        for one_operator_name in operator_name:
            operators.append(eval(one_operator_name))
        operator = tuple(operators)

    # make it an iterable
    if callable(operator):
        operator = (operator, )

    operations = []
    for one_operator in operator:
        on_axis = False
        keepdims = True

        try:
            args = getfullargspec(one_operator).args
        except TypeError:
            lg.debug('func ' + str(one_operator) + ' is not a Python '
                     'function')
        else:
            if 'axis' in args:
                on_axis = True

                if axis is None:
                    raise TypeError('You need to specify an axis if you '
                                    'use ' + one_operator.__name__ +
                                    ' (which applies to an axis)')

            if 'keepdims' in args or one_operator in NOKEEPDIM:
                keepdims = False

        operations.append({'name': one_operator.__name__,
                           'func': one_operator,
                           'on_axis': on_axis,
                           'keepdims': keepdims,
                           })

    output = data._copy()

    if axis is not None:
        idx_axis = data.index_of(axis)

    first_op = True
    for op in operations:
        #lg.info('running operator: ' + op['name'])
        func = op['func']

        if func == mode:
            func = lambda x, axis: mode(x, axis=axis)[0]

        for i in range(output.number_of('trial')):

            # don't copy original data, but use data if it's the first operation
            if first_op:
                x = data(trial=i)
            else:
                x = output(trial=i)

            if op['on_axis']:
                lg.debug('running ' + op['name'] + ' on ' + str(idx_axis))

                try:
                    if func == diff:
                        lg.debug('Diff has one-point of zero padding')
                        x = _pad_one_axis_one_value(x, idx_axis)
                    output.data[i] = func(x, axis=idx_axis)

                except IndexError:
                    raise ValueError('The axis ' + axis + ' does not '
                                     'exist in [' +
                                     ', '.join(list(data.axis.keys())) + ']')

            else:
                lg.debug('running ' + op['name'] + ' on each datapoint')
                output.data[i] = func(x)

        first_op = False

        if op['on_axis'] and not op['keepdims']:
            del output.axis[axis]

    return output

def get_descriptives(data):
    """Get mean, SD, and mean and SD of log values.

    Parameters
    ----------
    data : ndarray
        Data with segment as first dimension
        and all other dimensions raveled into second dimension.

    Returns
    -------
    dict of ndarray
        each entry is a 1-D vector of descriptives over segment dimension
    """
    output = {}
    dat_log = log(abs(data))
    output['mean'] = nanmean(data, axis=0)
    output['sd'] = nanstd(data, axis=0)
    output['mean_log'] = nanmean(dat_log, axis=0)
    output['sd_log'] = nanstd(dat_log, axis=0)

    return output

def _pad_one_axis_one_value(x, idx_axis):
    pad_width = [(0, 0)] * x.ndim
    pad_width[idx_axis] = (1, 0)
    return pad(x, pad_width=pad_width, mode='mean')

# additional operators
def dB(x):
    return 10 * log10(x)



"""Module to read open-ephys .continuous files

TODO
----
It assumes that the lenght of a block (i.e. record) is 1024 data points but this
might change in the future.
"""
from datetime import datetime
from logging import getLogger
from struct import unpack, calcsize
from xml.etree import ElementTree

from numpy import array, ones, empty, NaN

from .edf import _select_blocks

lg = getLogger(__name__)


HDR_LENGTH = 1024  # this is fixed
BLK_LENGTH = 1024  # this is currently fixed but might change in the future
BEG_BLK = '<qHH'  # time information and indices
DAT_FMT = '>' + BLK_LENGTH * 'h'  # actual data, big-endian
END_BLK = 10 * 'B'  # control values

BEG_BLK_SIZE = calcsize(BEG_BLK)
DAT_FMT_SIZE = calcsize(DAT_FMT)
BLK_SIZE = BEG_BLK_SIZE + DAT_FMT_SIZE + calcsize(END_BLK)


class OpenEphys:
    """Provide class OpenEphys, which can be used to read the folder
    containing Open-Ephys .continuous files

    Parameters
    ----------
    filename : Path
        Full path for the EDF file

    Attributes
    ----------
    channels : list of dict
        list of filenames referring to channels which are actually on disk
    blocks : 1D array
        length of each block, in samples (currently 1024)
    gain : 1D array
        gain to convert digital to microvolts (one value per channel)
    """
    def __init__(self, filename):
        self.filename = filename.resolve()
        self.openephys_file = filename / 'Continuous_Data.openephys'
        self.settings_xml = filename / 'settings.xml'

    def return_hdr(self):
        """Return the header for further use.

        Returns
        -------
        subj_id : str
            subject identification code
        start_time : datetime
            start time of the dataset
        s_freq : float
            sampling frequency
        chan_name : list of str
            list of all the channels
        n_samples : int
            number of samples in the dataset
        orig : dict
            currently empty for open-ephys
        """
        subj_id = self.filename.stem  # use directory name as subject name

        start_time = _read_date(self.settings_xml)

        s_freq, channels = _read_openephys(self.openephys_file)

        # only use channels that are actually in the folder
        chan_name = []
        self.channels = []
        gain = []
        for chan in channels:
            channel_filename = (self.filename / chan['filename'])
            if channel_filename.exists():
                chan_name.append(chan['name'])
                self.channels.append(channel_filename)
                gain.append(_check_header(channel_filename, s_freq))

            else:
                lg.warning(f'could not find {chan["filename"]} in {self.filename}')

        self.gain = array(gain)
        n_blocks, n_samples = _read_n_samples(self.channels[0])

        self.blocks = ones(n_blocks, dtype='int') * BLK_LENGTH

        orig = {}

        return subj_id, start_time, s_freq, chan_name, n_samples, orig

    def return_dat(self, chan, begsam, endsam):
        """Read the data for some/all of the channels

        Parameters
        ----------
        chan : list of int
            indices of the channels to read
        begsam : int
            start sample to read
        endsam : int
            end sample to read (exclusive)

        Returns
        -------
        2D array
            chan X samples recordings
        """
        dat = empty((len(chan), endsam - begsam))
        dat.fill(NaN)

        for i_chan, sel_chan in enumerate(chan):
            with self.channels[sel_chan].open('rb') as f:
                for i_dat, blk, i_blk in _select_blocks(self.blocks, begsam, endsam):
                    dat_in_rec = _read_block_continuous(f, blk)
                    dat[i_chan, i_dat[0]:i_dat[1]] = dat_in_rec[i_blk[0]:i_blk[1]]

        return dat * self.gain[chan, None]

    def return_markers(self):
        """Read the markers from the .events file

        TODO
        ----
        read markers from openephys
        """
        return []


def _read_block_continuous(f, i_block):
    """Read a single block / record completely

    Parameters
    ----------
    f : file handle
        handle to a file opened with 'rb'
    i_block : int
        index of the block to read

    Returns
    -------
    1D array
        data inside a block for one channel

    Notes
    -----
    It skips the timestamp information (it's assumed to be continuous) and the
    control characters. Maybe it might be useful to check the control
    characters but it will slow down the execution.
    """
    f.seek(HDR_LENGTH + i_block * BLK_SIZE + BEG_BLK_SIZE)
    v = unpack(DAT_FMT, f.read(DAT_FMT_SIZE))

    return array(v)


def _read_openephys(openephys_file):
    """Read the channel labels and their respective files from the
    'Continuous_Data.openephys' file

    Parameters
    ----------
    openephys_file : Path
        path to Continuous_Data.openephys inside the open-ephys folder

    Returns
    -------
    int
        sampling frequency
    list of dict
        list of channels containing the label, the filename and the gain
    """
    root = ElementTree.parse(openephys_file).getroot()

    channels = []
    for recording in root:
        s_freq = float(recording.attrib['samplerate'])
        for processor in recording:
            for channel in processor:
                channels.append(channel.attrib)

    return s_freq, channels


def _read_date(settings_file):
    """Get the data from the settings.xml file

    Parameters
    ----------
    settings_file : Path
        path to settings.xml inside open-ephys folder

    Returns
    -------
    datetime
        start time of the recordings

    Notes
    -----
    The start time is present in the header of each file. This might be useful
    if 'settings.xml' is not present.
    """
    root = ElementTree.parse(settings_file).getroot()
    for e0 in root:
        if e0.tag == 'INFO':
            for e1 in e0:
                if e1.tag == 'DATE':
                    break

    return datetime.strptime(e1.text, '%d %b %Y %H:%M:%S')


def _read_n_samples(channel_file):
    """Calculate the number of samples based on the file size

    Parameters
    ----------
    channel_file : Path
        path to single filename with the header

    Returns
    -------
    int
        number of blocks (i.e. records, in which the data is cut)
    int
        number of samples
    """
    n_blocks = int((channel_file.stat().st_size - HDR_LENGTH) / BLK_SIZE)
    n_samples = n_blocks * BLK_LENGTH
    return n_blocks, n_samples


def _read_header(filename):
    """Read the text header for each file

    Parameters
    ----------
    channel_file : Path
        path to single filename with the header

    Returns
    -------
    dict
        header
    """
    with filename.open('rb') as f:
        h = f.read(HDR_LENGTH).decode()

        header = {}
        for line in h.split('\n'):
            if '=' in line:
                key, value = line.split(' = ')
                key = key.strip()[7:]
                value = value.strip()[:-1]
                header[key] = value

    return header


def _check_header(channel_file, s_freq):
    """For each file, make sure that the header is consistent with the
    information in the text file.

    Parameters
    ----------
    channel_file : Path
        path to single filename with the header
    s_freq : int
        sampling frequency

    Returns
    -------
    int
        gain from digital to microvolts (the same information is stored in
        the Continuous_Data.openephys but I trust the header for each file more.
    """
    hdr = _read_header(channel_file)

    assert int(hdr['header_bytes']) == HDR_LENGTH
    assert int(hdr['blockLength']) == BLK_LENGTH
    assert int(hdr['sampleRate']) == s_freq

    return float(hdr['bitVolts'])



from copy import deepcopy
from os.path import dirname, split

from numpy import arange, atleast_2d, squeeze
try:
    from mne import morph_data, SourceEstimate
except ImportError:
    pass


class Morph:
    def __init__(self, from_surf, to_surf='fsaverage', smooth=None):
        self.from_surf = from_surf
        self.to_surf = to_surf
        self.smooth = smooth

    def __call__(self, data):

        output = deepcopy(data)

        from_surf_file = dirname(dirname(self.from_surf.surf_file))
        SUBJECTS_DIR, from_surf_name = split(from_surf_file)

        if 'lh' == self.from_surf.surf_file.stem:
            vertices = [arange(data.data[0].shape[0]), arange(0)]
        else:
            vertices = [arange(0), arange(data.data[0].shape[0])]

        try:
            stc = SourceEstimate(atleast_2d(data.data[0]).T, vertices=vertices,
                                 tstep=0, tmin=0)
        except NameError:
            raise ImportError('mne needs to be installed for this function')

        m = morph_data(from_surf_name, self.to_surf, stc,
                       subjects_dir=SUBJECTS_DIR, grade=None,
                       smooth=self.smooth, verbose=False)

        output.data[0] = squeeze(m.data, axis=1)
        output.axis['surf'][0] = arange(m.data.shape[0])

        return output



"""Module to keep track of the user-made annotations and sleep scoring.
"""
from logging import getLogger
from bisect import bisect_left
from itertools import groupby
from csv import reader, writer
from datetime import datetime, timedelta
from numpy import (allclose, around, asarray, clip, diff, isnan, logical_and, 
                   modf, nan)
from math import ceil, inf
from os.path import basename, splitext
from pathlib import Path
from re import search, sub
from scipy.io import loadmat
from xml.etree.ElementTree import Element, SubElement, tostring, parse
from xml.dom.minidom import parseString

try:
    from PyQt5.QtCore import Qt
    from PyQt5.QtWidgets import QProgressDialog
except ImportError:
    Qt = None
    QProgressDialog = None

from .. import __version__
from ..utils.exceptions import UnrecognizedFormat


lg = getLogger(__name__)
VERSION = '5'
DOMINO_STAGE_KEY = {'N1': 'NREM1',
                    'N2': 'NREM2',
                    'N3': 'NREM3',
                    'Re': 'REM',
                    'Wa': 'Wake',
                    'Ar': 'Artefact',
                    'A\n': 'Artefact'}

REMLOGIC_STAGE_KEY = {'SLEEP-S0': 'Wake',
                      'SLEEP-S1': 'NREM1',
                      'SLEEP-S2': 'NREM2',
                      'SLEEP-S3': 'NREM3',
                      'SLEEP-REM': 'REM',
                      'SLEEP-UNSCORED': 'Undefined'}

ALICE_STAGE_KEY = {'WK': 'Wake',
                   'N1': 'NREM1',
                   'N2': 'NREM2',
                   'N3': 'NREM3',
                   'EM': 'REM'}

SANDMAN_STAGE_KEY = {' 1': 'NREM1',
                     ' 2': 'NREM2',
                     ' 3': 'NREM3',
                     'em': 'REM',
                     'ke': 'Wake',
                     '1\t': 'NREM1',
                     '2\t': 'NREM2',
                     '3\t': 'NREM3',
                     'm\t': 'REM',
                     'e\t': 'Wake'}

COMPUMEDICS_STAGE_KEY = {'?': 'Unknown',
                         'W': 'Wake',
                         '1': 'NREM1',
                         '2': 'NREM2',
                         '3': 'NREM3',
                         'R': 'REM'}

FASST_STAGE_KEY = ['Wake',
                   'NREM1',
                   'NREM2',
                   'NREM3',
                   None,
                   'REM',
                   ]

PRANA_STAGE_KEY = {'0': 'Wake',
                   '1': 'NREM1',
                   '2': 'NREM2',
                   '3': 'NREM3',
                   '5': 'REM'}

PHYSIP_STAGE_KEY = {'0': 'NREM3',
                   '1': 'NREM2',
                   '2': 'NREM1',
                   '3': 'REM',
                   '4': 'Wake',
                   '5': 'Artefact'}

def parse_iso_datetime(date):
    try:
        return datetime.strptime(date, "%Y-%m-%dT%H:%M:%S")
    except ValueError:
        return datetime.strptime(date, "%Y-%m-%dT%H:%M:%S.%f")


def create_empty_annotations(xml_file, dataset):
    """Create an empty annotation file.

    Notes
    -----
    Dates are made time-zone unaware.
    """
    xml_file = Path(xml_file)
    root = Element('annotations')
    root.set('version', VERSION)

    info = SubElement(root, 'dataset')
    x = SubElement(info, 'filename')
    x.text = str(dataset.filename)
    x = SubElement(info, 'path')  # not to be relied on
    x.text = str(dataset.filename)
    x = SubElement(info, 'start_time')
    start_time = dataset.header['start_time'].replace(tzinfo=None)
    x.text = start_time.isoformat()

    first_sec = 0
    last_sec = int(dataset.header['n_samples'] /
                   dataset.header['s_freq'])  # in s

    x = SubElement(info, 'first_second')
    x.text = str(first_sec)
    x = SubElement(info, 'last_second')
    x.text = str(last_sec)

    xml = parseString(tostring(root))
    with xml_file.open('w') as f:
        f.write(xml.toxml())


def create_annotation(xml_file, from_fasst):
    """Create annotations by importing from FASST sleep scoring file.

    Parameters
    ----------
    xml_file : path to xml file
        annotation file that will be created
    from_fasst : path to FASST file
        .mat file containing the scores

    Returns
    -------
    instance of Annotations

    TODO
    ----
    Merge create_annotation and create_empty_annotations
    """
    xml_file = Path(xml_file)
    try:
        mat = loadmat(str(from_fasst), variable_names='D', struct_as_record=False,
                      squeeze_me=True)
    except ValueError:
        raise UnrecognizedFormat(str(from_fasst) + ' does not look like a FASST .mat file')

    D = mat['D']
    info = D.other.info
    score = D.other.CRC.score

    microsecond, second = modf(info.hour[2])
    start_time = datetime(*info.date, int(info.hour[0]), int(info.hour[1]),
                          int(second), int(microsecond * 1e6))
    first_sec = score[3, 0][0]
    last_sec = score[0, 0].shape[0] * score[2, 0]

    root = Element('annotations')
    root.set('version', VERSION)

    info = SubElement(root, 'dataset')
    x = SubElement(info, 'filename')
    x.text = D.other.info.fname
    x = SubElement(info, 'path')  # not to be relied on
    x.text = D.other.info.fname
    x = SubElement(info, 'start_time')
    x.text = start_time.isoformat()

    x = SubElement(info, 'first_second')
    x.text = str(int(first_sec))
    x = SubElement(info, 'last_second')
    x.text = str(int(last_sec))

    xml = parseString(tostring(root))
    with xml_file.open('w') as f:
        f.write(xml.toxml())

    annot = Annotations(xml_file)

    n_raters = score.shape[1]
    for i_rater in range(n_raters):
        rater_name = score[1, i_rater]
        epoch_length = int(score[2, i_rater])
        annot.add_rater(rater_name, epoch_length=epoch_length)

        for epoch_start, epoch in enumerate(score[0, i_rater]):
            if isnan(epoch):
                continue
            annot.set_stage_for_epoch(epoch_start * epoch_length,
                                      FASST_STAGE_KEY[int(epoch)], save=False)

    annot.save()

    return annot


class Annotations():
    """Class to return nicely formatted information from xml.

    Parameters
    ----------
    xml_file : path to xml file
        Annotation xml file
    """
    def __init__(self, xml_file, rater_name=None):

        self.xml_file = xml_file
        self.root = self.load()
        if rater_name is None:
            self.rater = self.root.find('rater')
        else:
            self.get_rater(rater_name)

    def load(self):
        """Load xml from file."""
        lg.info('Loading ' + str(self.xml_file))
        update_annotation_version(self.xml_file)

        xml = parse(self.xml_file)
        return xml.getroot()

    def save(self):
        """Save xml to file."""
        if self.rater is not None:
            self.rater.set('modified', datetime.now().isoformat())

        xml = parseString(tostring(self.root))
        with open(self.xml_file, 'w') as f:
            f.write(xml.toxml())

    @property
    def dataset(self):
        return self.root.find('dataset/path').text

    @property
    def start_time(self):
        return parse_iso_datetime(self.root.find('dataset/start_time').text)

    @property
    def first_second(self):
        return int(self.root.find('dataset/first_second').text)

    @property
    def last_second(self):
        return int(self.root.find('dataset/last_second').text)

    @property
    def current_rater(self):
        try:
            return self.rater.get('name')
        except AttributeError:
            raise IndexError('No rater in the annotations')

    @property
    def raters(self):
        return [rater.get('name') for rater in self.root.iter('rater')]

    @property
    def epoch_length(self):
        epoch = next(self.epochs)
        return around(epoch['end'] - epoch['start'])

    def get_rater(self, rater_name):
        # get xml root for one rater
        found = False

        for rater in self.root.iterfind('rater'):
            if rater.get('name') == rater_name:
                self.rater = rater
                found = True

        if not found:
            raise KeyError(rater_name + ' not in the list of raters (' +
                           ', '.join(self.raters) + ')')

    def add_rater(self, rater_name, epoch_length=30):
        if rater_name in self.raters:
            lg.warning('rater ' + rater_name + ' already exists, selecting it')
            self.get_rater(rater_name)
            return

        # add one rater + subtree
        rater = SubElement(self.root, 'rater')
        rater.set('name', rater_name)
        rater.set('created', datetime.now().isoformat())

        self.get_rater(rater_name)

        # create subtree
        SubElement(self.rater, 'bookmarks')
        SubElement(self.rater, 'events')
        SubElement(self.rater, 'stages')
        SubElement(self.rater, 'cycles')
        self.create_epochs(epoch_length=epoch_length)

        self.save()

    def rename_rater(self, name, new_name):
        """Rename event type."""
        for rater in self.root.iterfind('rater'):
            if rater.get('name') == name:
                rater.set('name', new_name)

        self.save()
    
    def remove_rater(self, rater_name):
        # remove one rater
        for rater in self.root.iterfind('rater'):
            if rater.get('name') == rater_name:

                # here we deal with the current rater
                if rater is self.rater:
                    all_raters = self.root.findall('rater')
                    if len(all_raters) == 1:
                        self.rater = None
                    else:
                        idx = all_raters.index(self.rater)
                        idx -= 1  # select the previous rater
                        if idx == -1:
                            idx = 1  # rater to delete is 0
                        self.rater = all_raters[idx]

                self.root.remove(rater)

        self.save()

    def import_staging(self, filename, source, rater_name, rec_start,
                       staging_start=None, poor=['Artefact'], as_qual=False):
        """Import staging from an external staging text file.

        Parameters
        ----------
        filename : str
            Staging file name.
        source : str
            Name of program where staging was made. One of 'domino', 'alice',
            'compumedics', 'sandman', 'remlogic'
        rater_name : str
            Rater name for imported staging.
        rec_start : datetime
            Date and time (year, month, day, hour, minute, second) of recording
            start. Year is ignored (New Year's Eve celebratory recordings
            unsupported.)
        staging_start : datetime (default: None)
            Date and time of staging start. For use when not provided in
            staging file.
        poor : list of str
            epochs with stage names in this list will be marked as Poor quality
        as_qual : bool
            if True, the staging only be used to mark quality, as per poor
        """            
        if rater_name not in self.raters:
            
            if as_qual:
                self.parent.statusBar.showMessage('Rater not found.')
                return
            
            self.add_rater(rater_name, epoch_length=30)

        self.get_rater(rater_name)
        
        if not as_qual:
            stages = self.rater.find('stages')
            # list is necessary so that it does not remove in place
            for s in list(stages):
                stages.remove(s)

        if source == 'sandman':
            encoding = 'ISO-8859-1'
        else:
            encoding = 'utf-8'

        with open(filename, 'r', encoding=encoding) as f:
            lines = f.readlines()
            stages = self.rater.find('stages')

            if source == 'domino':
                
                for i, line in enumerate(lines):
                    if line[0].isdigit():
                        first_line = i
                        break
                
                if lines[first_line].index(';') > 15:
                    idx_time = (11, 19)
                    idx_stage = (25, 26)
                    stage_key = PHYSIP_STAGE_KEY
                else:
                    idx_time = (0, 8)
                    idx_stage = (14, 16)
                    stage_key = DOMINO_STAGE_KEY
                
                stage_start = datetime.strptime(
                        lines[first_line][idx_time[0]:idx_time[1]], '%H:%M:%S')
                stage_day = int(lines[1][12:14])
                stage_month = int(lines[1][15:17])
                stage_start_for_delta = stage_start.replace(year=1999,
                                                            month=stage_month,
                                                            day=stage_day)
                rec_start_for_delta = rec_start.replace(year=1999)
                first_second = int((stage_start_for_delta -
                                    rec_start_for_delta).total_seconds())

            elif source == 'remlogic':
                stage_start = datetime.strptime(lines[14][:19],
                                                '%Y-%m-%dT%H:%M:%S')
                first_second = int((stage_start - rec_start).total_seconds())

                first_line = 14

                stage_key = {k[-2:]: v for k, v in REMLOGIC_STAGE_KEY.items()}
                idx_stage = (-6, -4)

            elif source == 'alice':
                stage_start = datetime.strptime(lines[1][2:13], '%I:%M:%S %p')
                dt = rec_start

                # best guess in absence of date
                if lines[1][11:13] == 'pm' and rec_start.hour < 12:
                    dt = rec_start - timedelta(days=1)
                elif lines[1][11:13] == 'am' and rec_start.hour > 12:
                    dt = rec_start + timedelta

                stage_start = stage_start.replace(year=dt.year,
                                                  month=dt.month,
                                                  day=dt.day)
                first_second = int((stage_start - rec_start).total_seconds())

                first_line = 1

                lines[-1] += '_' # to fill newline position
                stage_key = ALICE_STAGE_KEY
                idx_stage = (-3, -1)

            elif source == 'sandman':
                stage_start = datetime.strptime(lines[4][12:33],
                                                '%d/%m/%Y %I:%M:%S %p')
                first_second = int((stage_start - rec_start).total_seconds())

                first_line = 14

                stage_key = SANDMAN_STAGE_KEY
                idx_stage = (-14, -12)

            elif source == 'compumedics':
                if staging_start is None:
                    first_second = 0
                else:
                    first_second = int((
                            staging_start - rec_start).total_seconds())

                first_line = 0

                stage_key = COMPUMEDICS_STAGE_KEY
                idx_stage = (0, 1)
                
            elif source == 'prana':
                stage_start = datetime.strptime(lines[5][:11], '%d %H:%M:%S')
                
                # best guess in absence of date
                dt = rec_start
                if stage_start.hour > 12 and rec_start.hour < 12:
                    dt = rec_start - timedelta(days=1)
                elif stage_start.hour < 12 and rec_start.hour > 12:
                    dt = rec_start + timedelta(days=1)                                
                stage_start = stage_start.replace(year=dt.year,
                                                  month=dt.month,
                                                  day=dt.day)                
                first_second = int((stage_start - rec_start).total_seconds())
                
                first_line = 5
                
                stage_key = PRANA_STAGE_KEY
                
                spacer = next(i for i, j in enumerate(lines[5][30:]) \
                              if j.strip())
                idx_stage = (30 + spacer, 30 + spacer + 1)

            else:
                raise ValueError('Unknown source program for staging file')

            lg.info('Time offset: ' + str(first_second) + ' sec')

            if as_qual:
                
                for i, one_line in enumerate(lines[first_line:]):                     
                    
                    if one_line[idx_stage[0]:-1] in poor:
                        epoch_beg = first_second + (i * 30)
                        
                        try:
                            self.set_stage_for_epoch(epoch_beg, 'Poor', 
                                                     attr='quality', 
                                                     save=False)
                        except KeyError:
                            return 1
            
            else:
            
                for i, one_line in enumerate(lines[first_line:]):
                    epoch = SubElement(stages, 'epoch')
    
                    start_time = SubElement(epoch, 'epoch_start')
                    epoch_beg = first_second + (i * 30)
                    start_time.text = str(epoch_beg)
    
                    end_time = SubElement(epoch, 'epoch_end')
                    end_time.text = str(epoch_beg + 30)
    
                    epoch_stage = SubElement(epoch, 'stage')
    
                    try:
                        key = one_line[idx_stage[0]:idx_stage[1]]
                        one_stage = stage_key[key]
    
                    except KeyError:
                        one_stage = 'Unknown'
                        lg.info('Stage not recognized: ' + key)
    
                    epoch_stage.text = one_stage
    
                    quality = SubElement(epoch, 'quality')
                    if one_stage in poor:
                        quality.text = 'Poor'
                    else:
                        quality.text = 'Good'

        self.save()
    
    def add_bookmark(self, name, time, chan=''):
        """Add a new bookmark

        Parameters
        ----------
        name : str
            name of the bookmark
        time : (float, float)
            float with start and end time in s

        Raises
        ------
        IndexError
            When there is no selected rater
        """
        try:
            bookmarks = self.rater.find('bookmarks')
        except AttributeError:
            raise IndexError('You need to have at least one rater')
        new_bookmark = SubElement(bookmarks, 'bookmark')
        bookmark_name = SubElement(new_bookmark, 'bookmark_name')
        bookmark_name.text = name
        bookmark_time = SubElement(new_bookmark, 'bookmark_start')
        bookmark_time.text = str(time[0])
        bookmark_time = SubElement(new_bookmark, 'bookmark_end')
        bookmark_time.text = str(time[1])

        if isinstance(chan, (tuple, list)):
            chan = ', '.join(chan)
        event_chan = SubElement(new_bookmark, 'bookmark_chan')
        event_chan.text = chan

        self.save()

    def remove_bookmark(self, name=None, time=None, chan=None):
        """if you call it without arguments, it removes ALL the bookmarks."""
        bookmarks = self.rater.find('bookmarks')

        for m in bookmarks:

            bookmark_name = m.find('bookmark_name').text
            bookmark_start = float(m.find('bookmark_start').text)
            bookmark_end = float(m.find('bookmark_end').text)
            bookmark_chan = m.find('bookmark_chan').text
            if bookmark_chan is None:  # xml doesn't store empty string
                bookmark_chan = ''

            if name is None:
                name_cond = True
            else:
                name_cond = bookmark_name == name

            if time is None:
                time_cond = True
            else:
                time_cond = (time[0] <= bookmark_end and
                             time[1] >= bookmark_start)

            if chan is None:
                chan_cond = True
            else:
                chan_cond = bookmark_chan == chan

            if name_cond and time_cond and chan_cond:
                bookmarks.remove(m)

        self.save()

    def get_bookmarks(self, time=None, chan=None):
        """
        Raises
        ------
        IndexError
            When there is no selected rater
        """
        # get bookmarks inside window
        try:
            bookmarks = self.rater.find('bookmarks')
        except AttributeError:
            raise IndexError('You need to have at least one rater')

        mrks = []
        for m in bookmarks:

            bookmark_start = float(m.find('bookmark_start').text)
            bookmark_end = float(m.find('bookmark_end').text)
            bookmark_chan = m.find('bookmark_chan').text
            if bookmark_chan is None:  # xml doesn't store empty string
                bookmark_chan = ''

            if time is None:
                time_cond = True
            else:
                time_cond = (time[0] <= bookmark_end and
                             time[1] >= bookmark_start)

            if chan is None:
                chan_cond = True
            else:
                chan_cond = bookmark_chan == chan

            if time_cond and chan_cond:
                one_mrk = {'name': m.find('bookmark_name').text,
                           'start': bookmark_start,
                           'end': bookmark_end,
                           'chan': bookmark_chan.split(', '),  # always a list
                           }
                mrks.append(one_mrk)

        return mrks

    @property
    def event_types(self):
        """
        Raises
        ------
        IndexError
            When there is no selected rater
        """
        try:
            events = self.rater.find('events')
        except AttributeError:
            raise IndexError('You need to have at least one rater')

        return [x.get('type') for x in events]

    def add_event_type(self, name):
        """
        Raises
        ------
        IndexError
            When there is no selected rater
        """
        if name in self.event_types:
            lg.info('Event type ' + name + ' exists already.')
            return

        events = self.rater.find('events')
        new_event_type = SubElement(events, 'event_type')
        new_event_type.set('type', name)
        self.save()

    def remove_event_type(self, name):
        """Remove event type based on name."""

        if name not in self.event_types:
            lg.info('Event type ' + name + ' was not found.')

        events = self.rater.find('events')

        # list is necessary so that it does not remove in place
        for e in list(events):
            if e.get('type') == name:
                events.remove(e)

        self.save()
        
    def rename_event_type(self, name, new_name):
        """Rename event type."""

        if name not in self.event_types:
            lg.info('Event type ' + name + ' was not found.')

        events = self.rater.find('events')

        for e in list(events):
            if e.get('type') == name:
                e.set('type', new_name)

        self.save()

    def add_event(self, name, time, chan=''):
        """Add event to annotations file.
        Parameters
        ----------
        name : str
            Event type name.
        time : tuple/list of float
            Start and end times of event, in seconds from recording start.
        chan : str or list of str, optional
            Channel or channels associated with event.
        Raises
        ------
        IndexError
            When there is no rater / epochs at all
        """
        if name not in self.event_types:
            self.add_event_type(name)

        events = self.rater.find('events')
        pattern = "event_type[@type='" + name + "']"
        event_type = events.find(pattern)

        new_event = SubElement(event_type, 'event')
        event_start = SubElement(new_event, 'event_start')
        event_start.text = str(time[0])
        event_end = SubElement(new_event, 'event_end')
        event_end.text = str(time[1])

        if isinstance(chan, (tuple, list)):
            chan = ', '.join(chan)
        event_chan = SubElement(new_event, 'event_chan')
        event_chan.text = chan

        event_qual = SubElement(new_event, 'event_qual')
        event_qual.text = 'Good' # if the event was marked, it's probably
        # because the signal quality is good; anyway, it gets checked against
        # the epoch quality in get_events (JOB)

        self.save()

    def remove_event(self, name=None, time=None, chan=None):
        """get events inside window."""
        events = self.rater.find('events')
        if name is not None:
            pattern = "event_type[@type='" + name + "']"
        else:
            pattern = "event_type"

        if chan is not None:
            if isinstance(chan, (tuple, list)):
                chan = ', '.join(chan)

        for e_type in list(events.iterfind(pattern)):

            for e in e_type:

                event_start = float(e.find('event_start').text)
                event_end = float(e.find('event_end').text)
                event_chan = e.find('event_chan').text

                if time is None:
                    time_cond = True
                else:
                    time_cond = allclose(time[0], event_start) and allclose(
                            time[1], event_end)

                if chan is None:
                    chan_cond = True
                else:
                    chan_cond = event_chan == chan

                if time_cond and chan_cond:
                    e_type.remove(e)

        self.save()

    def get_events(self, name=None, time=None, chan=None, stage=None,
                   qual=None):
        """Get list of events in the file.

        Parameters
        ----------
        name : str, optional
            name of the event of interest
        time : tuple of two float, optional
            start and end time of the period of interest
        chan : tuple of str, optional
            list of channels of interests
        stage : tuple of str, optional
            list of stages of interest
        qual : str, optional
            epoch signal qualifier (Good or Poor)
        Returns
        -------
        list of dict
            where each dict has 'name' (name of the event), 'start' (start
            time), 'end' (end time), 'chan' (channels of interest, can be
            empty), 'stage', 'quality' (signal quality)

        Raises
        ------
        IndexError
            When there is no rater / epochs at all
        """
        # get events inside window
        events = self.rater.find('events')
        if name is not None:
            pattern = "event_type[@type='" + name + "']"
        else:
            pattern = "event_type"

        if chan is not None:
            if isinstance(chan, (tuple, list)):
                if chan[0] is not None:
                    chan = ', '.join(chan)
                else:
                    chan = None

        if stage or qual:
            ep_starts = [x['start'] for x in self.epochs]
            if stage:
                ep_stages = [x['stage'] for x in self.epochs]
            if qual:
                ep_quality = [x['quality'] for x in self.epochs]

        ev = []
        for e_type in events.iterfind(pattern):

            event_name = e_type.get('type')

            for e in e_type:

                event_start = float(e.find('event_start').text)
                event_end = float(e.find('event_end').text)
                event_chan = e.find('event_chan').text
                event_qual = e.find('event_qual').text
                if event_chan is None:  # xml doesn't store empty string
                    event_chan = ''

                if stage or qual:
                    pos = bisect_left(ep_starts, event_start)
                    if pos == len(ep_starts):
                        pos -= 1
                    elif event_start != ep_starts[pos]:
                        pos -= 1

                if stage is None:
                    stage_cond = True
                else:
                    ev_stage = ep_stages[pos]
                    stage_cond = ev_stage in stage

                if qual is None:
                    qual_cond = True
                else:
                    ev_qual = ep_quality[pos]
                    qual_cond = ev_qual == qual

                if time is None:
                    time_cond = True
                else:
                    time_cond = time[0] <= event_end and time[1] >= event_start

                if chan is None:
                    chan_cond = True
                else:
                    chan_cond = event_chan == chan

                if time_cond and chan_cond and stage_cond and qual_cond:
                    one_ev = {'name': event_name,
                              'start': event_start,
                              'end': event_end,
                              'chan': event_chan.split(', '),  # always a list
                              'stage': '',
                              'quality': event_qual
                              }
                    if stage is not None:
                        one_ev['stage'] = ev_stage
                    ev.append(one_ev)

        return ev

    def create_epochs(self, epoch_length=30, first_second=None):
        """Create epochs in annotation file.
        Parameters
        ----------
        epoch_length : int
            duration in seconds of each epoch
        first_second : int, optional
            Time, in seconds from record start, at which the epochs begin
        """
        lg.info('creating epochs of length ' + str(epoch_length))
        if first_second is None:
            first_second = self.first_second
        last_sec = ceil((self.last_second - first_second) /
                        epoch_length) * epoch_length

        stages = self.rater.find('stages')
        for epoch_beg in range(first_second, last_sec, epoch_length):
            epoch = SubElement(stages, 'epoch')

            start_time = SubElement(epoch, 'epoch_start')
            start_time.text = str(epoch_beg)

            end_time = SubElement(epoch, 'epoch_end')
            end_time.text = str(epoch_beg + epoch_length)

            stage = SubElement(epoch, 'stage')
            stage.text = 'Unknown'

            quality = SubElement(epoch, 'quality')
            quality.text = 'Good'

    @property
    def epochs(self):
        """Get epochs as generator

        Returns
        -------
        list of dict
            each epoch is defined by start_time and end_time (in s in reference
            to the start of the recordings) and a string of the sleep stage,
            and a string of the signal quality.
            If you specify stages_of_interest, only epochs belonging to those
            stages will be included (can be an empty list).

        Raises
        ------
        IndexError
            When there is no rater / epochs at all
        """
        if self.rater is None:
            raise IndexError('You need to have at least one rater')

        for one_epoch in self.rater.iterfind('stages/epoch'):
            epoch = {'start': int(one_epoch.find('epoch_start').text),
                     'end': int(one_epoch.find('epoch_end').text),
                     'stage': one_epoch.find('stage').text,
                     'quality': one_epoch.find('quality').text
                     }
            yield epoch

    def get_epochs(self, time=None, stage=None, qual=None,
                   chan=None, name=None):
        """Get list of events in the file.

        Parameters
        ----------
        time : tuple of two float, optional
            start and end time of the period of interest
        stage : tuple of str, optional
            list of stages of interest
        qual : str, optional
            epoch signal qualifier (Good or Poor)
        chan : None
            placeholder, to maintain format similar to get_events
        name : None
            placeholder, to maintain format similar to get_events
        Returns
        -------
        list of dict
            where each dict has 'start' (start time), 'end' (end time),
            'stage', 'qual' (signal quality)
        """
        time_cond = True
        stage_cond = True
        qual_cond = True
        valid = []

        for ep in self.epochs:
            if stage:
                stage_cond = ep['stage'] in stage
            if qual:
                qual_cond = ep['quality'] == qual
            if time:
                time_cond = time[0] <= ep['start'] and time[1] >= ep['end']
            if stage_cond and qual_cond and time_cond:
                valid.append(ep)

        return valid

    def get_epoch_start(self, window_start):
        """ Get the position (seconds) of the nearest epoch.

        Parameters
        ----------
        window_start : float
            Position of the current window (seconds)

        Returns
        -------
        float
            Position (seconds) of the nearest epoch.
        """
        epoch_starts = [x['start'] for x in self.epochs]
        idx = asarray([abs(window_start - x) for x in epoch_starts]).argmin()

        return epoch_starts[idx]

    def get_stage_for_epoch(self, epoch_start, window_length=None,
                            attr='stage'):
        """Return stage for one specific epoch.

        Parameters
        ----------
        id_epoch : str
            index of the epoch
        attr : str, optional
            'stage' or 'quality'

        Returns
        -------
        stage : str
            description of the stage.
        """
        for epoch in self.epochs:
            if epoch['start'] == epoch_start:
                return epoch[attr]

            if window_length is not None:
                epoch_length = epoch['end'] - epoch['start']
                if logical_and(window_length < epoch_length,
                               0 <= \
                               (epoch_start - epoch['start']) < epoch_length):
                    return epoch[attr]

    def time_in_stage(self, name, attr='stage'):
        """Return time (in seconds) in the selected stage.

        Parameters
        ----------
        name : str
            one of the sleep stages, or qualifiers
        attr : str, optional
            either 'stage' or 'quality'

        Returns
        -------
        int
            time spent in one stage/qualifier, in seconds.

        """
        return sum(x['end'] - x['start'] for x in self.epochs
                   if x[attr] == name)

    def set_stage_for_epoch(self, epoch_start, name, attr='stage', save=True):
        """Change the stage for one specific epoch.

        Parameters
        ----------
        epoch_start : int
            start time of the epoch, in seconds
        name : str
            description of the stage or qualifier.
        attr : str, optional
            either 'stage' or 'quality'
        save : bool
            whether to save every time one epoch is scored

        Raises
        ------
        KeyError
            When the epoch_start is not in the list of epochs.
        IndexError
            When there is no rater / epochs at all

        Notes
        -----
        In the GUI, you want to save as often as possible, even if it slows
        down the program, but it's the safer option. But if you're converting
        a dataset, you want to save at the end. Do not forget to save!
        """
        if self.rater is None:
            raise IndexError('You need to have at least one rater')

        for one_epoch in self.rater.iterfind('stages/epoch'):
            if int(one_epoch.find('epoch_start').text) == epoch_start:
                one_epoch.find(attr).text = name
                if save:
                    self.save()
                return

        raise KeyError('epoch starting at ' + str(epoch_start) + ' not found')

    def set_cycle_mrkr(self, epoch_start, end=False):
        """Mark epoch start as cycle start or end.

        Parameters
        ----------
        epoch_start: int
            start time of the epoch, in seconds
        end : bool
            If True, marked as cycle end; otherwise, marks cycle start
        """
        if self.rater is None:
            raise IndexError('You need to have at least one rater')

        bound = 'start'
        if end:
            bound = 'end'

        for one_epoch in self.rater.iterfind('stages/epoch'):
            if int(one_epoch.find('epoch_start').text) == epoch_start:
                cycles = self.rater.find('cycles')
                name = 'cyc_' + bound
                new_bound = SubElement(cycles, name)
                new_bound.text = str(int(epoch_start))
                self.save()
                return

        raise KeyError('epoch starting at ' + str(epoch_start) + ' not found')

    def remove_cycle_mrkr(self, epoch_start):
        """Remove cycle marker at epoch_start.

        Parameters
        ----------
        epoch_start: int
            start time of epoch, in seconds
        """
        if self.rater is None:
            raise IndexError('You need to have at least one rater')
        cycles = self.rater.find('cycles')
        for one_mrkr in cycles.iterfind('cyc_start'):
            if int(one_mrkr.text) == epoch_start:
                cycles.remove(one_mrkr)
                self.save()
                return

        for one_mrkr in cycles.iterfind('cyc_end'):
            if int(one_mrkr.text) == epoch_start:
                cycles.remove(one_mrkr)
                self.save()
                return

        raise KeyError('cycle marker at ' + str(epoch_start) + ' not found')

    def clear_cycles(self):
        """Remove all cycle markers in current rater."""
        if self.rater is None:
            raise IndexError('You need to have at least one rater')

        cycles = self.rater.find('cycles')
        for cyc in list(cycles):
            cycles.remove(cyc)

        self.save()

    def get_cycles(self):
        """Return the cycle start and end times.

        Returns
        -------
        list of tuple of float
            start and end times for each cycle, in seconds from recording start
            and the cycle index starting at 1
        """
        cycles = self.rater.find('cycles')

        if not cycles:
            return None

        starts = sorted(
                [float(mrkr.text) for mrkr in cycles.findall('cyc_start')])
        ends = sorted(
                [float(mrkr.text) for mrkr in cycles.findall('cyc_end')])
        cyc_list = []

        if not starts or not ends:
            return None

        if all(i < starts[0] for i in ends):
            raise ValueError('First cycle has no start.')

        for (this_start, next_start) in zip(starts, starts[1:] + [inf]):
            # if an end is smaller than the next start, make it the end
            # otherwise, the next_start is the end
            end_between_starts = [end for end in ends \
                                  if this_start < end <= next_start]

            if len(end_between_starts) > 1:
                raise ValueError('Found more than one cycle end for same '
                                 'cycle')

            if end_between_starts:
                one_cycle = (this_start, end_between_starts[0])
            else:
                one_cycle = (this_start, next_start)

            if one_cycle[1] == inf:
                raise ValueError('Last cycle has no end.')

            cyc_list.append(one_cycle)

        output = []
        for i, j in enumerate(cyc_list):
            cyc = j[0], j[1], i + 1
            output.append(cyc)

        return output

    def switch(self):
        """Obtain switch parameter, ie number of times the stage shifts."""
        stag_to_int = {'NREM1': 1, 'NREM2': 2, 'NREM3': 3, 'REM': 5, 'Wake': 0}
        hypno = [stag_to_int[x['stage']] for x in self.get_epochs() \
                 if x['stage'] in stag_to_int.keys()]
        
        return sum(asarray(diff(hypno), dtype=bool))
    
    def slp_frag(self):
        """Obtain sleep fragmentation parameter, ie number of stage shifts to 
        a lighter stage."""
        epochs = self.get_epochs()
        stage_int = {'Wake': 0, 'NREM1': 1, 'NREM2': 2, 'NREM3': 3, 'REM': 2}
        
        hypno_str = [x['stage'] for x in epochs \
                     if x['stage'] in stage_int.keys()]
        hypno_int = [stage_int[x] for x in hypno_str]
        frag = sum(asarray(clip(diff(hypno_int), a_min=None, a_max=0), 
                           dtype=bool))
            
        # N3 to REM doesn't count
        n3_to_rem = 0
        for i, j in enumerate(hypno_str):
            if j == 'NREM3':
                if hypno_str[i + 1] == 'REM':
                    n3_to_rem += 1
        
        return frag - n3_to_rem
    
    def latency_to_consolidated(self, lights_off, duration=5, 
                                stage=['NREM2', 'NREM3']):
        """Find latency to the first period of uninterrupted 'stage'.
        
        Parameters
        ----------
        lights_off : float
            lights off time, in seconds form recording start
        duration : float
            duration of uninterrupted period, in minutes
        stage : list of str
            target stage(s)
            
        Returns
        -------
        float
            latency to the start of the consolidated period, in minutes
        """
        epochs = self.get_epochs()
        
        if len(stage) > 1:
            for ep in epochs:
                if ep['stage'] in stage:
                    ep['stage'] = 'target'
            stage = ['target']
            
        hypno = [x['stage'] for x in epochs]        
        groups = groupby(hypno)
        runs = [(stag, sum(1 for _ in group)) for stag, group in groups]
        
        idx_start = 0
        for one_stage, n in runs:
            if (one_stage in stage) and n >= duration * 60 / self.epoch_length:
                break
            idx_start += n
            
        if idx_start < len(hypno):
            latency = (epochs[idx_start]['start'] - lights_off) / 60 
        else:
            latency = nan
        
        return latency
    
    def export(self, file_to_export, xformat='csv'):
        """Export epochwise annotations to csv file.

        Parameters
        ----------
        file_to_export : path to file
            file to write to
        """
        if 'csv' == xformat:
        
            with open(file_to_export, 'w', newline='') as f:
                csv_file = writer(f)
                csv_file.writerow(['Wonambi v{}'.format(__version__)])
                csv_file.writerow(('clock start time', 'start', 'end',
                                   'stage'))
    
                for epoch in self.epochs:
                    epoch_time = (self.start_time +
                                  timedelta(seconds=epoch['start']))
                    csv_file.writerow((epoch_time.strftime('%H:%M:%S'),
                                       epoch['start'],
                                       epoch['end'],
                                       epoch['stage']))
                    
        if 'remlogic' in xformat:
            
            columns = 'Time [hh:mm:ss]\tEvent\tDuration[s]\n'
            if 'remlogic_fr' == xformat:
                columns = 'Heure [hh:mm:ss]\tEvénement\tDurée[s]\n'
                
            patient_id = splitext(basename(self.dataset))[0]
            rec_date = self.start_time.strftime('%d/%m/%Y')
            stkey = {v:k for k, v in REMLOGIC_STAGE_KEY.items()}
            stkey['Artefact'] = 'SLEEP-UNSCORED'
            stkey['Unknown'] = 'SLEEP-UNSCORED'
            stkey['Movement'] = 'SLEEP-UNSCORED'
            
            with open(file_to_export, 'w') as f:
                f.write('RemLogic Event Export\n')
                f.write('Patient:\t' + patient_id + '\n')
                f.write('Patient ID:\t' + patient_id + '\n')
                f.write('Recording Date:\t' + rec_date + '\n')
                f.write('\n')
                f.write('Events Included:\n')
                
                for i in sorted(set([stkey[x['stage']] for x in self.epochs])):
                    f.write(i + '\n')
                
                f.write('\n')
                f.write(columns)
                
                for epoch in self.epochs:
                    epoch_time = (self.start_time +
                                  timedelta(seconds=epoch['start']))
                    f.write((epoch_time.strftime('%Y-%m-%dT%H:%M:%S.000000') + 
                             '\t' + 
                             stkey[epoch['stage']] + 
                             '\t' + 
                             str(self.epoch_length) + 
                             '\n'))

    def export_sleep_stats(self, filename, lights_off, lights_on):
        """Create CSV with sleep statistics.

        Parameters
        ----------
        filename: str
            Filename for csv export
        lights_off: float
            Initial time when sleeper turns off the light (or their phone) to
            go to sleep, in seconds from recording start
        lights_on: float
            Final time when sleeper rises from bed after sleep, in seconds from
            recording start

        Returns
        -------
        float or None
            If there are no epochs scored as sleep, returns None. Otherwise,
            returns the sleep onset latency, for testing purposes.
        """
        epochs = self.get_epochs()
        ep_starts = [i['start'] for i in epochs]
        hypno = [i['stage'] for i in epochs]
        n_ep_per_min = 60 / self.epoch_length

        first = {}
        latency = {}
        for stage in ['NREM1', 'NREM2', 'NREM3', 'REM']:
            first[stage] = next(((i, j) for i, j in enumerate(epochs) if \
                                j['stage'] == stage), None)
            if first[stage] is not None:
                latency[stage] = (first[stage][1]['start'] - 
                       lights_off) / 60
            else:
                first[stage] = nan
                latency[stage] = nan

        idx_loff = asarray([abs(x - lights_off) for x in ep_starts]).argmin()
        idx_lon = asarray([abs(x - lights_on) for x in ep_starts]).argmin()
        duration = {}
        for stage in ['NREM1', 'NREM2', 'NREM3', 'REM', 'Wake', 'Movement',
                      'Artefact']:
            duration[stage] = hypno[idx_loff:idx_lon].count(
                    stage) / n_ep_per_min

        slp_onset = sorted(first.values(), key=lambda x: x[1]['start'])[0]
        wake_up = next((len(epochs) - i, j) for i, j in enumerate(
                epochs[::-1]) if j['stage'] in ['NREM1', 'NREM2', 'NREM3',
                                                'REM'])
        total_dark_time = (lights_on - lights_off) / 60
        #slp_period_time = (wake_up[1]['start'] - slp_onset[1]['start']) / 60
        slp_onset_lat = (slp_onset[1]['start'] - lights_off) / 60
        waso = hypno[slp_onset[0]:wake_up[0]].count('Wake') / n_ep_per_min
        wake = waso + slp_onset_lat
        total_slp_period = sum((waso, duration['NREM1'], duration['NREM2'],
                                  duration['NREM3'], duration['REM']))
        total_slp_time = total_slp_period - waso
        slp_eff = total_slp_time / total_dark_time
        switch = self.switch()
        slp_frag = self.slp_frag()
        
        dt_format = '%d/%m/%Y %H:%M:%S'
        loff_str = (self.start_time + timedelta(seconds=lights_off)).strftime(
                dt_format)
        lon_str = (self.start_time + timedelta(seconds=lights_on)).strftime(
                dt_format)
        slp_onset_str = (self.start_time + timedelta(
                seconds=slp_onset[1]['start'])).strftime(dt_format)
        wake_up_str = (self.start_time + timedelta(
                seconds=wake_up[1]['start'])).strftime(dt_format)
        
        slcnrem5 = self.latency_to_consolidated(lights_off, duration=5, 
                                                stage=['NREM2', 'NREM3'])
        slcnrem10 = self.latency_to_consolidated(lights_off, duration=10, 
                                                 stage=['NREM2', 'NREM3'])
        slcn35 = self.latency_to_consolidated(lights_off, duration=5, 
                                              stage=['NREM3'])
        slcn310 = self.latency_to_consolidated(lights_off, duration=10, 
                                               stage=['NREM3'])
        
        with open(filename, 'w', newline='') as f:
            lg.info('Writing to ' + str(filename))
            cf = writer(f)
            cf.writerow(['Wonambi v{}'.format(__version__)])            
            cf.writerow(['Variable', 'Acronym', 
                         'Unit 1', 'Value 1', 
                         'Unit 2', 'Value 2', 
                         'Formula'])
            cf.writerow(['Lights off', 'LOFF', 
                         'dd/mm/yyyy HH:MM:SS', loff_str, 
                         'seconds from recording start', lights_off, 
                         'marker'])
            cf.writerow(['Lights on', 'LON', 
                         'dd/mm/yyyy HH:MM:SS', lon_str, 
                         'seconds from recording start', lights_on, 
                         'marker'])
            cf.writerow(['Sleep onset', 'SO', 
                         'dd/mm/yyyy HH:MM:SS', slp_onset_str, 
                         'seconds from recording start', slp_onset[1]['start'], 
                         'first sleep epoch (N1 or N2) - LOFF'])
            cf.writerow(['Time of last awakening', '',
                         'dd/mm/yyyy HH:MM:SS', wake_up_str,
                         'seconds from recording start', wake_up[1]['start'],
                         'end time of last epoch of N1, N2, N3 or REM'])
            cf.writerow(['Total dark time (Time in bed)', 'TDT (TIB)', 
                         'Epochs', total_dark_time * n_ep_per_min, 
                         'Minutes', total_dark_time, 
                         'LON - LOFF'])
            cf.writerow(['Sleep latency', 'SL', 
                         'Epochs', slp_onset_lat * n_ep_per_min, 
                         'Minutes', slp_onset_lat, 
                         'LON - SO'])
            cf.writerow(['Wake', 'W', 
                         'Epochs', wake * n_ep_per_min, 
                         'Minutes', wake,
                         'total wake duration between LOFF and LON'])
            cf.writerow(['Wake after sleep onset', 'WASO', 
                         'Epochs', waso * n_ep_per_min, 
                         'Minutes', waso, 
                         'W - SL'])
            cf.writerow(['N1 duration', '',
                         'Epochs', duration['NREM1'] * n_ep_per_min,
                         'Minutes', duration['NREM1'],
                         'total N1 duration between LOFF and LON'])
            cf.writerow(['N2 duration', '',
                         'Epochs', duration['NREM2'] * n_ep_per_min,
                         'Minutes', duration['NREM2'],
                         'total N2 duration between LOFF and LON'])
            cf.writerow(['N3 duration', '',
                         'Epochs', duration['NREM3'] * n_ep_per_min, 
                         'Minutes', duration['NREM3'],
                         'total N3 duration between LOFF and LON'])
            cf.writerow(['REM duration', '', 
                         'Epochs', duration['REM'] * n_ep_per_min,
                         'Minutes', duration['REM'],
                         'total REM duration between LOFF and LON'])            
            cf.writerow(['Artefact duration', '', 
                         'Epochs', 
                         duration['Artefact'] * n_ep_per_min,
                         'Minutes', duration['Artefact'],
                         'total Artefact duration between LOFF and LON'])
            cf.writerow(['Movement duration', '', 
                         'Epochs', 
                         duration['Movement'] * n_ep_per_min,
                         'Minutes', duration['Movement'],
                         'total Movement duration between LOFF and LON'])
            cf.writerow(['Total sleep period', 'TSP', 
                         'Epochs', total_slp_period * n_ep_per_min, 
                         'Minutes', total_slp_period,
                         'WASO + N1 + N2 + N3 + REM'])
            cf.writerow(['Total sleep time', 'TST', 
                         'Epochs', total_slp_time * n_ep_per_min, 
                         'Minutes', total_slp_time, 
                         'N1 + N2 + N3 + REM'])
            cf.writerow(['Sleep efficiency', 'SE', 
                         '%', slp_eff * 100, 
                         '', '',
                         'TST / TDT'])
            cf.writerow(['W % TSP', '',
                         '%', waso * 100 / total_slp_period,
                         '', '',
                         'WASO / TSP'])
            cf.writerow(['N1 % TSP', '',
                         '%', duration['NREM1'] * 100 / total_slp_period,
                         '', '',
                         'N1 / TSP'])
            cf.writerow(['N2 % TSP', '',
                         '%', duration['NREM2'] * 100 / total_slp_period,
                         '', '',
                         'N2 / TSP'])
            cf.writerow(['N3 % TSP', '',
                         '%', duration['NREM3'] * 100 / total_slp_period,
                         '', '',
                         'N3 / TSP'])
            cf.writerow(['REM % TSP', '',
                         '%', duration['REM'] * 100 / total_slp_period,
                         '', '',
                         'REM / TSP'])
            cf.writerow(['N1 % TST', '',
                         '%', duration['NREM1'] * 100 / total_slp_time,
                         '', '',
                         'N1 / TST'])
            cf.writerow(['N2 % TST', '',
                         '%', duration['NREM2'] * 100 / total_slp_time,
                         '', '',
                         'N2 / TST'])
            cf.writerow(['N3 % TST', '',
                         '%', duration['NREM3'] * 100 / total_slp_time,
                         '', '',
                         'N3 / TST'])
            cf.writerow(['REM % TST', '',
                         '%', duration['REM'] * 100 / total_slp_time,
                         '', '',
                         'REM / TST'])
            cf.writerow(['Switch', '',
                         'N', switch,
                         '', '', 
                         'number of stage shifts'])
            cf.writerow(['Switch %', '',
                         '% epochs', 
                         switch * 100 / total_slp_period / n_ep_per_min,
                         '% minutes', switch * 100 / total_slp_period,
                         'switch / TSP'])
            cf.writerow(['Sleep fragmentation', '',
                         'N', slp_frag,
                         '', '', 
                         ('number of shifts to a lighter stage '
                          '(W > N1 > N2 > N3; W > N1 > REM)')])
            cf.writerow(['Sleep fragmentation index', 'SFI', 
                         '% epochs', 
                         slp_frag * 100 / total_slp_time / n_ep_per_min, 
                         '% minutes', slp_frag * 100 / total_slp_time,
                         'sleep fragmentation / TST'])
            cf.writerow(['Sleep latency to N1', 'SLN1', 
                         'Epochs', latency['NREM1'] * n_ep_per_min, 
                         'Minutes', latency['NREM1'],
                         'first N1 epoch - LOFF'])
            cf.writerow(['Sleep latency to N2', 'SLN2', 
                         'Epochs', latency['NREM2'] * n_ep_per_min, 
                         'Minutes', latency['NREM2'],
                         'first N2 epoch - LOFF'])
            cf.writerow(['Sleep latency to N3', 'SLN3', 
                         'Epochs', latency['NREM3'] * n_ep_per_min, 
                         'Minutes', latency['NREM3'],
                         'first N3 epoch - LOFF'])
            cf.writerow(['Sleep latency to REM', 'SLREM', 
                         'Epochs', latency['REM'] * n_ep_per_min, 
                         'Minutes', latency['REM'],
                         'first REM epoch - LOFF'])
            cf.writerow(['Sleep latency to consolidated NREM, 5 min', 
                         'SLCNREM5', 
                         'Epochs', slcnrem5 * n_ep_per_min, 
                         'Minutes', slcnrem5,
                         ('start of first uninterrupted 5-minute period of '
                          'N2 and/or N3 - LOFF')])
            cf.writerow(['Sleep latency to consolidated NREM, 10 min', 
                         'SLCNREM10', 
                         'Epochs', slcnrem10 * n_ep_per_min, 
                         'Minutes', slcnrem10,
                         ('start of first uninterrupted 10-minute period of '
                          'N2 and/or N3 - LOFF')])
            cf.writerow(['Sleep latency to consolidated N3, 5 min', 'SLCN35', 
                         'Epochs', slcn35 * n_ep_per_min, 
                         'Minutes', slcn35,
                         ('start of first uninterrupted 5-minute period of '
                          'N3 - LOFF')])
            cf.writerow(['Sleep latency to consolidated N3, 10 min', 'SLCN310', 
                         'Epochs', slcn310 * n_ep_per_min, 
                         'Minutes', slcn310,
                         ('start of first uninterrupted 10-minute period of '
                          'N3 - LOFF')])

        return slp_onset_lat, waso, total_slp_time # for testing

    def export_events(self, filename, evt_type):
        """Export events to CSV
        
        Parameters
        ----------
        filename : str
            path of export file
        evt_type : list of str
            event types to export
        """
        filename = splitext(filename)[0] + '.csv'
        headings_row = ['Index',
                       'Start time',
                       'End time',
                       'Stitches',
                       'Stage',
                       'Cycle',
                       'Event type',
                       'Channel']
        
        events = []
        for et in evt_type:
            events.extend(self.get_events(name=et))
            
        events = sorted(events, key=lambda evt: evt['start'])
        
        if events is None:
            lg.info('No events found.')
            return
        
        with open(filename, 'w', newline='') as f:
            lg.info('Writing to ' + str(filename))
            csv_file = writer(f)
            csv_file.writerow(['Wonambi v{}'.format(__version__)])
            csv_file.writerow(headings_row)
            
            for i, ev in enumerate(events):
                csv_file.writerow([i + 1,
                                   ev['start'],
                                   ev['end'],
                                   0,
                                   ev['stage'],
                                   '',
                                   ev['name'],
                                   ', '.join(ev['chan']),
                                   ])
            
    def import_events(self, source_file, parent=None):
        """Import events from Wonambi CSV event export and write to annot.
        
        Parameters
        ----------
        source_file : str
            path to file CSV file
        parent : QWidget
            for GUI progress bar
        """
        events = []
        
        with open(source_file, 'r', encoding='utf-8') as csvfile:
            csv_reader = reader(csvfile, delimiter=',')
            
            for row in csv_reader:
                try:
                    int(row[0])
                    one_ev = {'name': row[6],
                              'start': float(row[1]),
                              'end': float(row[2]),
                              'chan': row[7].split(', '),  # always a list
                              'stage': row[4],
                              'quality': 'Good'
                              }
                    events.append(one_ev)
                    
                except ValueError:
                    continue
                
            if parent is not None:
                progress = QProgressDialog('Saving events', 'Abort',
                                   0, len(events) - 1, parent)
                progress.setWindowModality(Qt.ApplicationModal)
                
            for i, one_ev in enumerate(events):
                self.add_event(one_ev['name'],
                                (one_ev['start'], one_ev['end']),
                                chan=one_ev['chan'])
                
                if parent is not None:
                    progress.setValue(i)                    
                    if progress.wasCanceled():
                        return

        if parent is not None:
            progress.close()
        

def update_annotation_version(xml_file):
    """Update the fields that have changed over different versions.

    Parameters
    ----------
    xml_file : path to file
        xml file with the sleep scoring

    Notes
    -----
    new in version 4: use 'marker_name' instead of simply 'name' etc

    new in version 5: use 'bookmark' instead of 'marker'
    """
    with open(xml_file, 'r') as f:
        s = f.read()

    m = search('<annotations version="([0-9]*)">', s)
    current = int(m.groups()[0])

    if current < 4:
        s = sub('<marker><name>(.*?)</name><time>(.*?)</time></marker>',
                 '<marker><marker_name>\g<1></marker_name><marker_start>\g<2></marker_start><marker_end>\g<2></marker_end><marker_chan/></marker>',
                 s)

    if current < 5:
        s = s.replace('marker', 'bookmark')

        # note indentation
        s = sub('<annotations version="[0-9]*">',
                '<annotations version="5">', s)
        with open(xml_file, 'w') as f:
            f.write(s)
